#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from copy import deepcopy
from typing import Dict, List

import numpy as np
import pandas as pd

from ..utils.textureTools import (coord2index, get_neighbour_direction,
                                  get_value, is_list_all_none)


def get_matrix(roi_only: np.array,
                     levels: np.ndarray) -> float:
    """Computes Neighbouring grey level dependence matrix.
    This matrix refers to "Neighbouring grey level dependence based features" (ID = REK0)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        roi_only_int (ndarray): Smallest box containing the ROI, with the imaging data ready
            for texture analysis computations. Voxels outside the ROI are
            set to NaNs.
        levels (ndarray or List): Vector containing the quantized gray-levels
            in the tumor region (or reconstruction ``levels`` of quantization).

    Returns:
        ndarray: Array of neighbouring grey level dependence matrix of ``roi_only``.

    """
    roi_only = roi_only.copy()

    # PRELIMINARY
    level_temp = np.max(levels)+1
    roi_only[np.isnan(roi_only)] = level_temp
    levels = np.append(levels, level_temp)
    dim = np.shape(roi_only)
    if np.size(dim) == 2:
        np.append(dim, 1)

    q_2 = np.reshape(roi_only, np.prod(dim), order='F').astype("int")

    # QUANTIZATION EFFECTS CORRECTION (M. Vallieres)
    # In case (for example) we initially wanted to have 64 levels, but due to
    # quantization, only 60 resulted.
    # q_s = round(levels*adjust)/adjust;
    # q_2 = round(q_2*adjust)/adjust;
    q_s = levels.copy()

    # EL NAQA CODE
    q_3 = q_2*0
    lqs = np.size(q_s)
    for k in range(1, lqs+1):
        q_3[q_2 == q_s[k-1]] = k

    q_3 = np.reshape(q_3, dim, order='F')

    # Min dependence = 0, Max dependence = 26; So 27 columns
    ngldm = np.zeros((lqs, 27))
    for i in range(1, dim[0]+1):
        i_min = max(1, i-1)
        i_max = min(i+1, dim[0])
        for j in range(1, dim[1]+1):
            j_min = max(1, j-1)
            j_max = min(j+1, dim[1])
            for k in range(1, dim[2]+1):
                k_min = max(1, k-1)
                k_max = min(k+1, dim[2])
                val_q3 = q_3[i-1, j-1, k-1]
                count = 0
                for I2 in range(i_min, i_max+1):
                    for J2 in range(j_min, j_max+1):
                        for K2 in range(k_min, k_max+1):
                            if (I2 == i) & (J2 == j) & (K2 == k):
                                continue
                            else:
                                # a = 0
                                if (val_q3 - q_3[I2-1, J2-1, K2-1] == 0):
                                    count += 1

                ngldm[val_q3-1, count] = ngldm[val_q3-1, count] + 1

    # Last column was for the NaN voxels, to be removed
    ngldm = np.delete(ngldm, -1, 0)
    stop = np.nonzero(np.sum(ngldm, 0))[0][-1]
    ngldm = np.delete(ngldm, range(stop+1, np.shape(ngldm)[1]+1), 1)

    return ngldm

def extract_all(vol: np.ndarray) -> Dict :
    """Compute NGLDM features

    Args:
        vol (np.ndarray): volume with discretised intensities as 3D numpy array (x, y, z)
        method (str, optional): Either 'old' (deprecated) or 'new' (faster) method.

    Raises:
        ValueError: Ngldm should either be calculated using the faster \"new\" method, or the slow \"old\" method.

    Returns:
        dict: Dictionary of NGTDM features.
    """
    
    ngldm = get_ngldm_features(vol=vol)

    return ngldm


def get_ngldm_features(vol: np.ndarray,
                       ngldm_spatial_method: str="3d",
                       ngldm_diff_lvl: float=0.0,
                       ngldm_dist: float=1.0) -> Dict:
    """Extract neighbouring grey level dependence matrix-based features from the intensity roi mask.
    These features refer to "Neighbouring grey level dependence based features" (ID = REK0) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:

        vol (ndarray): volume with discretised intensities as 3D numpy array (x, y, z).
        intensity_range (ndarray): range of potential discretised intensities,
                         provided as a list: [minimal discretised intensity, maximal discretised intensity].
                         If one or both values are unknown, replace the respective values with np.nan.
        ngldm_spatial_method(str): spatial method which determines the way neighbouring grey level dependence
                              matrices are calculated and how features are determined. One of "2d", "2.5d" or "3d".
        ngldm_diff_lvl (float): also called coarseness. Coarseness determines which intensity 
                        differences are allowed for intensities to be considered similar. Typically 0, and
                        changing discretisation levels may have the same effect as increasing
                        the coarseness parameter.
        ngldm_dist (float): the chebyshev distance that forms a local neighbourhood around a center voxel.

    Returns:
        dict: dictionary with feature values.
    """
    if type(ngldm_spatial_method) is not list:
        ngldm_spatial_method = [ngldm_spatial_method]

    if type(ngldm_diff_lvl) is not list:
        ngldm_diff_lvl = [ngldm_diff_lvl]

    if type(ngldm_dist) is not list:
        ngldm_dist = [ngldm_dist]

    # Get the roi in tabular format
    img_dims = vol.shape
    index_id = np.arange(start=0, stop=vol.size)
    coords = np.unravel_index(indices=index_id, shape=img_dims)
    df_img = pd.DataFrame({"index_id": index_id,
                           "g": np.ravel(vol),
                           "x": coords[0],
                           "y": coords[1],
                           "z": coords[2],
                           "roi_int_mask": np.ravel(np.isfinite(vol))})

    # Generate an empty feature list
    feat_list = []

    # Iterate over spatial arrangements
    for ii_spatial in ngldm_spatial_method:

        # Iterate over difference levels
        for ii_diff_lvl in ngldm_diff_lvl:

            # Iterate over distances
            for ii_dist in ngldm_dist:

                # Initiate list of ngldm objects
                ngldm_list = []

                # Perform 2D analysis
                if ii_spatial.lower() in ["2d", "2.5d"]:

                    # Iterate over slices
                    for ii_slice in np.arange(0, img_dims[2]):

                        # Add ngldm matrices to list
                        ngldm_list += [GreyLevelDependenceMatrix(distance=int(ii_dist), diff_lvl=ii_diff_lvl,
                                                                 spatial_method=ii_spatial.lower(), img_slice=ii_slice)]

                # Perform 3D analysis
                elif ii_spatial.lower() == "3d":

                    # Add ngldm matrices to list
                    ngldm_list += [GreyLevelDependenceMatrix(distance=int(ii_dist), diff_lvl=ii_diff_lvl,
                                                             spatial_method=ii_spatial.lower(), img_slice=None)]
                else:
                    raise ValueError("Spatial methods for ngldm should be \"2d\", \"2.5d\" or \"3d\".")

                # Calculate ngldm matrices
                for ngldm in ngldm_list:
                    ngldm.calculate_ngldm_matrix(df_img=df_img, img_dims=img_dims)

                # Merge matrices according to the given method
                upd_list = combine_ngldm_matrices(ngldm_list=ngldm_list, spatial_method=ii_spatial.lower())

                # Calculate features
                feat_run_list = []
                for ngldm in upd_list:
                    feat_run_list += [ngldm.compute_ngldm_features()]

                # Average feature values
                feat_list += [pd.concat(feat_run_list, axis=0).mean(axis=0, skipna=True).to_frame().transpose()]

    # Merge feature tables into a single dictionary
    df_feat = pd.concat(feat_list, axis=1).to_dict(orient="records")[0]

    return df_feat


def combine_ngldm_matrices(ngldm_list: List,
                           spatial_method: str) -> List:
    """Function to merge neighbouring grey level dependence matrices prior to feature calculation.

    Args:
        ngldm_list (List): list of GreyLevelDependenceMatrix objects.
        spatial_method (str): spatial method which determines the way neighbouring grey level
                              dependence matrices are calculated and how features are determined.
                              One of "2d", "2.5d" or "3d".

    Returns:
        List: list of one or more merged GreyLevelDependenceMatrix objects.
    """
    # Initiate empty list
    use_list = []

    if spatial_method == "2d":
        # Average features over slice: maintain original ngldms

        # Make copy of ngldm_list
        use_list = []
        for ngldm in ngldm_list:
            use_list += [ngldm.copy()]

    elif spatial_method in ["2.5d", "3d"]:
        # Merge all ngldms into a single representation

        # Select all matrices within the slice
        sel_matrix_list = []
        for ngldm_id in np.arange(len(ngldm_list)):
            sel_matrix_list += [ngldm_list[ngldm_id].matrix]

        # Check if any matrix has been created
        if is_list_all_none(sel_matrix_list):
            # No matrix was created
            use_list += [GreyLevelDependenceMatrix(distance=ngldm_list[0].distance, diff_lvl=ngldm_list[0].diff_lvl,
                                                   spatial_method=spatial_method, img_slice=None, matrix=None, n_v=0.0)]
        else:
            # Merge neighbouring grey level difference matrices
            merge_ngldm = pd.concat(sel_matrix_list, axis=0)
            merge_ngldm = merge_ngldm.groupby(by=["i", "j"]).sum().reset_index()

            # Update the number of voxels
            merge_n_v = 0.0
            for ngldm_id in np.arange(len(ngldm_list)):
                merge_n_v += ngldm_list[ngldm_id].n_v

            # Create new neighbouring grey level difference matrix
            use_list += [GreyLevelDependenceMatrix(distance=ngldm_list[0].distance, diff_lvl=ngldm_list[0].diff_lvl,
                                                   spatial_method=spatial_method, img_slice=None, matrix=merge_ngldm, n_v=merge_n_v)]
    else:
        use_list = None

    # Return to new ngldm list to calling function
    return use_list


class GreyLevelDependenceMatrix:

    def __init__(self,
                 distance: float,
                 diff_lvl: float,
                 spatial_method: str,
                 img_slice: np.ndarray=None,
                 matrix: np.ndarray=None,
                 n_v: float=None) ->  None:
        """Initialising function for a new neighbouring grey level dependence ``matrix``.
    
        Args:
            distance (float): chebyshev ``distance`` used to determine the local neighbourhood.
            diff_lvl (float): coarseness parameter which determines which intensities are considered similar.
            spatial_method (str): spatial method used to calculate the ngldm: 2d, 2.5d or 3d
            img_slice (ndarray): corresponding slice index (only if the ngldm corresponds to a 2d image slice)
            matrix (ndarray): the actual ngldm in sparse format (row, column, count)
            n_v (float): the number of voxels in the volume
        """

        # Distance used
        self.distance = distance
        self.diff_lvl = diff_lvl

        # Slice for which the current matrix is extracted
        self.img_slice = img_slice

        # Spatial analysis method (2d, 2.5d, 3d)
        self.spatial_method = spatial_method

        # Place holders
        self.matrix = matrix
        self.n_v = n_v

    def copy(self):
        """Returns a copy of the GreyLevelDependenceMatrix object."""
        return deepcopy(self)

    def set_empty(self):
        """Creates an empty GreyLevelDependenceMatrix"""
        self.n_v = 0
        self.matrix = None

    def calculate_ngldm_matrix(self,
                               df_img: pd.DataFrame,
                               img_dims: int):
        """
        Function that calculates an ngldm for the settings provided during initialisation and the input image.
        
        Args:
            df_img (pd.DataFrame): data table containing image intensities, x, y and z coordinates,
                and mask labels corresponding to voxels in the volume.
            img_dims (int): dimensions of the image volume
        """

        # Check if the input image and roi exist
        if df_img is None:
            self.set_empty()
            return

        # Check if the roi contains any masked voxels. If this is not the case, don't construct the ngldm.
        if not np.any(df_img.roi_int_mask):
            self.set_empty()
            return

        if self.spatial_method == "3d":
            # Set up neighbour vectors
            nbrs = get_neighbour_direction(d=self.distance, distance="chebyshev", centre=False, complete=True, dim3=True)

            # Set up work copy
            df_ngldm = deepcopy(df_img)
        elif self.spatial_method in ["2d", "2.5d"]:
            # Set up neighbour vectors
            nbrs = get_neighbour_direction(d=self.distance, distance="chebyshev", centre=False, complete=True, dim3=False)

            # Set up work copy
            df_ngldm = deepcopy(df_img[df_img.z == self.img_slice])
            df_ngldm["index_id"] = np.arange(0, len(df_ngldm))
            df_ngldm["z"] = 0
            df_ngldm = df_ngldm.reset_index(drop=True)
        else:
            raise ValueError("The spatial method for neighbouring grey level dependence matrices should be one of \"2d\", \"2.5d\" or \"3d\".")

        # Set grey level of voxels outside ROI to NaN
        df_ngldm.loc[df_ngldm.roi_int_mask == False, "g"] = np.nan

        # Update number of voxels for current iteration
        self.n_v = np.sum(df_ngldm.roi_int_mask.values)

        # Initialise sum of grey levels and number of neighbours
        df_ngldm["occur"] = 0.0
        df_ngldm["n_nbrs"] = 0.0

        for k in range(0, np.shape(nbrs)[1]):
            # Determine potential transitions from valid voxels
            df_ngldm["to_index"] = coord2index(x=df_ngldm.x.values + nbrs[2, k],
                                               y=df_ngldm.y.values + nbrs[1, k],
                                               z=df_ngldm.z.values + nbrs[0, k],
                                               dims=img_dims)

            # Get grey level value from transitions
            df_ngldm["to_g"] = get_value(x=df_ngldm.g.values, index=df_ngldm.to_index.values)

            # Determine which voxels have valid neighbours
            sel_index = np.isfinite(df_ngldm.to_g)

            # Determine co-occurrence within diff_lvl
            df_ngldm.loc[sel_index, "occur"] += ((np.abs(df_ngldm.to_g - df_ngldm.g)[sel_index]) <= self.diff_lvl) * 1

        # Work with voxels within the intensity roi
        df_ngldm = df_ngldm[df_ngldm.roi_int_mask]

        # Drop superfluous columns
        df_ngldm = df_ngldm.drop(labels=["index_id", "x", "y", "z", "to_index", "to_g", "roi_int_mask"], axis=1)

        # Sum s over voxels
        df_ngldm = df_ngldm.groupby(by=["g", "occur"]).size().reset_index(name="n")

        # Rename columns
        df_ngldm.columns = ["i", "j", "s"]

        # Add one to dependency count as features are not defined for k=0
        df_ngldm.j += 1.0

        # Add matrix to object
        self.matrix = df_ngldm

    def compute_ngldm_features(self) -> pd.DataFrame:
        """Computes neighbouring grey level dependence matrix features for the current neighbouring grey level dependence matrix.
        
        Returns:
            pandas data frame: with values for each feature.
        """
        # Create feature table
        feat_names = ["Fngl_lde",
                      "Fngl_hde",
                      "Fngl_lgce",
                      "Fngl_hgce",
                      "Fngl_ldlge",
                      "Fngl_ldhge",
                      "Fngl_hdlge",
                      "Fngl_hdhge",
                      "Fngl_glnu",
                      "Fngl_glnu_norm",
                      "Fngl_dcnu",
                      "Fngl_dcnu_norm",
                      "Fngl_gl_var",
                      "Fngl_dc_var",
                      "Fngl_dc_entr",
                      "Fngl_dc_energy"]
        df_feat = pd.DataFrame(np.full(shape=(1, len(feat_names)), fill_value=np.nan))
        df_feat.columns = feat_names

        # Don't return data for empty slices or slices without a good matrix
        if self.matrix is None:
            # Update names
            # df_feat.columns += self.parse_feature_names()
            return df_feat
        elif len(self.matrix) == 0:
            # Update names
            # df_feat.columns += self.parse_feature_names()
            return df_feat

        # Dependence count dataframe
        df_sij = deepcopy(self.matrix)
        df_sij.columns = ("i", "j", "sij")

        # Sum over grey levels
        df_si = df_sij.groupby(by="i")["sij"].agg(np.sum).reset_index().rename(columns={"sij": "si"})

        # Sum over dependence counts
        df_sj = df_sij.groupby(by="j")["sij"].agg(np.sum).reset_index().rename(columns={"sij": "sj"})

        # Constant definitions
        n_s = np.sum(df_sij.sij) * 1.0  # Number of neighbourhoods considered
        n_v = self.n_v  # Number of voxels

        ###############################################
        # ngldm features
        ###############################################

        # Low dependence emphasis
        df_feat.loc[0, "Fngl_lde"] = np.sum(df_sj.sj / df_sj.j ** 2.0) / n_s

        # High dependence emphasis
        df_feat.loc[0, "Fngl_hde"] = np.sum(df_sj.sj * df_sj.j ** 2.0) / n_s

        # Grey level non-uniformity
        df_feat.loc[0, "Fngl_glnu"] = np.sum(df_si.si ** 2.0) / n_s

        # Grey level non-uniformity, normalised
        df_feat.loc[0, "Fngl_glnu_norm"] = np.sum(df_si.si ** 2.0) / n_s ** 2.0

        # Dependence count non-uniformity
        df_feat.loc[0, "Fngl_dcnu"] = np.sum(df_sj.sj ** 2.0) / n_s

        # Dependence count non-uniformity, normalised
        df_feat.loc[0, "Fngl_dcnu_norm"] = np.sum(df_sj.sj ** 2.0) / n_s ** 2.0

        # Dependence count percentage
        # df_feat.loc[0, "ngl_dc_perc"] = n_s / n_v

        # Low grey level count emphasis
        df_feat.loc[0, "Fngl_lgce"] = np.sum(df_si.si / df_si.i ** 2.0) / n_s

        # High grey level count emphasis
        df_feat.loc[0, "Fngl_hgce"] = np.sum(df_si.si * df_si.i ** 2.0) / n_s

        # Low dependence low grey level emphasis
        df_feat.loc[0, "Fngl_ldlge"] = np.sum(df_sij.sij / (df_sij.i * df_sij.j) ** 2.0) / n_s

        # Low dependence high grey level emphasis
        df_feat.loc[0, "Fngl_ldhge"] = np.sum(df_sij.sij * df_sij.i ** 2.0 / df_sij.j ** 2.0) / n_s

        # High dependence low grey level emphasis
        df_feat.loc[0, "Fngl_hdlge"] = np.sum(df_sij.sij * df_sij.j ** 2.0 / df_sij.i ** 2.0) / n_s

        # High dependence high grey level emphasis
        df_feat.loc[0, "Fngl_hdhge"] = np.sum(df_sij.sij * df_sij.i ** 2.0 * df_sij.j ** 2.0) / n_s

        # Grey level variance
        mu = np.sum(df_sij.sij * df_sij.i) / n_s
        df_feat.loc[0, "Fngl_gl_var"] = np.sum((df_sij.i - mu) ** 2.0 * df_sij.sij) / n_s
        del mu

        # Dependence count variance
        mu = np.sum(df_sij.sij * df_sij.j) / n_s
        df_feat.loc[0, "Fngl_dc_var"] = np.sum((df_sij.j - mu) ** 2.0 * df_sij.sij) / n_s
        del mu

        # Dependence count entropy
        df_feat.loc[0, "Fngl_dc_entr"] = - np.sum(df_sij.sij * np.log2(df_sij.sij / n_s)) / n_s

        # Dependence count energy
        df_feat.loc[0, "Fngl_dc_energy"] = np.sum(df_sij.sij ** 2.0) / (n_s ** 2.0)

        # Update names
        # df_feat.columns += self.parse_feature_names()

        return df_feat

    def parse_feature_names(self):
        """
        Adds additional settings-related identifiers to each feature.
        Not used currently, as the use of different settings for the
        neighbouring grey level dependence matrix is not supported.
        """
        parse_str = ""

        # Add distance
        parse_str += "_d" + str(np.round(self.distance, 1))

        # Add difference level
        parse_str += "_a" + str(np.round(self.diff_lvl, 0))

        # Add spatial method
        if self.spatial_method is not None:
            parse_str += "_" + self.spatial_method

        return parse_str

def get_dict(vol: np.ndarray) -> dict:
    """
    Extract neighbouring grey level dependence matrix-based features from the intensity roi mask.

    Args:
        vol (ndarray): volume with discretised intensities as 3D numpy array (x, y, z)
    
    Returns:
        dict: dictionary with feature values

    """
    ngldm_dict =  get_ngldm_features(vol, intensity_range=[np.nan, np.nan])
    return ngldm_dict

def lde(ngldm_dict: np.ndarray)-> float:
    """
    Computes low dependence emphasis feature.
    This feature refers to "Fngl_lde" (ID = SODN) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: low depence emphasis value

    """
    return ngldm_dict["Fngl_lde"]

def hde(ngldm_dict: np.ndarray)-> float:
    """
    Computes high dependence emphasis feature.
    This feature refers to "Fngl_hde" (ID = IMOQ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: high depence emphasis value

    """
    return ngldm_dict["Fngl_hde"]

def lgce(ngldm_dict: np.ndarray)-> float:
    """
    Computes low grey level count emphasis feature.
    This feature refers to "Fngl_lgce" (ID = TL9H) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: low grey level count emphasis value

    """
    return ngldm_dict["Fngl_lgce"]

def hgce(ngldm_dict: np.ndarray)-> float:
    """
    Computes high grey level count emphasis feature.
    This feature refers to "Fngl_hgce" (ID = OAE7) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: high grey level count emphasis value

    """
    return ngldm_dict["Fngl_hgce"]

def ldlge(ngldm_dict: np.ndarray)-> float:
    """
    Computes low dependence low grey level emphasis feature.
    This feature refers to "Fngl_ldlge" (ID = EQ3F) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: low dependence low grey level emphasis value

    """
    return ngldm_dict["Fngl_ldlge"]

def ldhge(ngldm_dict: np.ndarray)-> float:
    """
    Computes low dependence high grey level emphasis feature.
    This feature refers to "Fngl_ldhge" (ID = JA6D) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: low dependence high grey level emphasis value

    """
    return ngldm_dict["Fngl_ldhge"]

def hdlge(ngldm_dict: np.ndarray)-> float:
    """
    Computes high dependence low grey level emphasis feature.
    This feature refers to "Fngl_hdlge" (ID = NBZI) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: high dependence low grey level emphasis value

    """
    return ngldm_dict["Fngl_hdlge"]

def hdhge(ngldm_dict: np.ndarray)-> float:
    """
    Computes high dependence high grey level emphasis feature.
    This feature refers to "Fngl_hdhge" (ID = 9QMG) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: high dependence high grey level emphasis value

    """
    return ngldm_dict["Fngl_hdhge"]

def glnu(ngldm_dict: np.ndarray)-> float:
    """
    Computes grey level non-uniformity feature.
    This feature refers to "Fngl_glnu" (ID = FP8K) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: grey level non-uniformity value

    """
    return ngldm_dict["Fngl_glnu"]

def glnu_norm(ngldm_dict: np.ndarray)-> float:
    """
    Computes grey level non-uniformity normalised feature.
    This feature refers to "Fngl_glnu_norm" (ID = 5SPA) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: grey level non-uniformity normalised value

    """
    return ngldm_dict["Fngl_glnu_norm"]

def dcnu(ngldm_dict: np.ndarray)-> float:
    """
    Computes dependence count non-uniformity feature.
    This feature refers to "Fngl_dcnu" (ID = Z87G) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: dependence count non-uniformity value

    """
    return ngldm_dict["Fngl_dcnu"]

def dcnu_norm(ngldm_dict: np.ndarray)-> float:
    """
    Computes dependence count non-uniformity normalised feature.
    This feature refers to "Fngl_dcnu_norm" (ID = OKJI) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: dependence count non-uniformity normalised value

    """
    return ngldm_dict["Fngl_dcnu_norm"]

def gl_var(ngldm_dict: np.ndarray)-> float:
    """
    Computes grey level variance feature.
    This feature refers to "Fngl_gl_var" (ID = 1PFV) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: grey level variance value

    """
    return ngldm_dict["Fngl_gl_var"]

def dc_var(ngldm_dict: np.ndarray)-> float:
    """
    Computes dependence count variance feature.
    This feature refers to "Fngl_dc_var" (ID = DNX2) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: dependence count variance value

    """
    return ngldm_dict["Fngl_dc_var"]

def dc_entr(ngldm_dict: np.ndarray)-> float:
    """
    Computes dependence count entropy feature.
    This feature refers to "Fngl_dc_entr" (ID = FCBV) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: dependence count entropy value

    """
    return ngldm_dict["Fngl_dc_entr"]

def dc_energy(ngldm_dict: np.ndarray)-> float:
    """
    Computes dependence count energy feature.
    This feature refers to "Fngl_dc_energy" (ID = CAS9) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngldm (ndarray): array of neighbouring grey level dependence matrix
    
    Returns:
        float: dependence count energy value

    """
    return ngldm_dict["Fngl_dc_energy"]



import random
from typing import Dict, List

import numpy as np
import pandas as pd


class DataCleaner:
    """
    Class that will clean features of the csv by removing features with too many missing values,
    too little variation, too many missing values per sample, too little variation per sample,
    and imputing missing values.
    """
    def __init__(self, df_features: pd.DataFrame, type: str = "continuous"):
        """
        Constructor of the class DataCleaner

        Args:
            df_features (pd.DataFrame): Table of features.
            type (str): Type of variable: "continuous", "hcategorical" or "icategorical". Defaults to "continuous".
        """
        self.df_features = df_features
        self.type = type
    
    def __update_df_features(self, var_of_type: List[str], flag_var_out: List[bool]) -> List[str]:
        """
        Updates the variable table by deleting the features that are not in the variable of type

        Args:
            var_of_type (List[str]): List of variable names.
            flag_var_out (List[bool]): List of variables to flag out.
        
        Returns:
            List[str]: List of variable names that were not flagged out.
        """
        var_to_delete = np.delete(var_of_type, [i for i, v in enumerate(flag_var_out) if not v])
        var_of_type = np.delete(var_of_type, [i for i, v in enumerate(flag_var_out) if v])
        self.df_features = self.df_features.drop(var_to_delete, axis=1)
        return var_of_type

    def cut_off_missing_per_sample(self, var_of_type: List[str], missing_cutoff : float = 0.25) -> None:
        """
        Removes observations/samples with more than ``missing_cutoff`` missing features.

        Args:
            var_of_type (List[str]): List of variable names.
            missing_cutoff (float): Maximum percentage cut-offs of missing features per sample. Defaults to 25%.
        
        Returns:
            None.
        """
        # Initialization
        n_observation, n_features = self.df_features.shape
        empty_vec = np.zeros(n_observation, dtype=int)
        data = self.df_features[var_of_type]
        empty_vec += data.isna().sum(axis=1).values
        
        # Gathering results
        ind_obs_out = np.where(((empty_vec/n_features) > missing_cutoff) == True)
        self.df_features = self.df_features.drop(self.df_features.index[ind_obs_out])
    
    def cut_off_missing_per_feature(self, var_of_type: List[str], missing_cutoff : float = 0.1) -> List[str]:
        """
        Removes features with more than ``missing_cutoff`` missing patients.

        Args:
            var_of_type (list): List of variable names.
            missing_cutoff (float): maximal percentage cut-offs of missing patient samples per variable.
        
        Returns:
            List[str]: List of variable names that were not flagged out.
        """
        flag_var_out = (((self.df_features[var_of_type].isna().sum()) / self.df_features.shape[0]) > missing_cutoff)
        return self.__update_df_features(var_of_type, flag_var_out)

    def cut_off_variation(self, var_of_type: List[str], cov_cutoff : float = 0.1) -> List[str]:
        """
        Removes features with a coefficient of variation (cov) less than ``cov_cutoff``.

        Args:
            var_of_type (list): List of variable names.
            cov_cutoff (float): minimal coefficient of variation cut-offs over samples per variable. Defaults to 10%.
        
        Returns:
            List[str]: List of variable names that were not flagged out.
        """
        eps = np.finfo(np.float32).eps
        cov_df_features = (self.df_features[var_of_type].std(skipna=True) / self.df_features[var_of_type].mean(skipna=True))
        flag_var_out = cov_df_features.abs().add(eps) < cov_cutoff
        return self.__update_df_features(var_of_type, flag_var_out)
    
    def impute_missing(self, var_of_type: List[str], imputation_method : str = "mean") -> None:
        """
        Imputes missing values of the features of type.

        Args:
            var_of_type (list): List of variable names.
            imputation_method (str): Method of imputation. Can be "mean", "median", "mode" or "random".
                For "random" imputation, a seed can be provided by adding the seed value after the method 
                name, for example "random42".
        
        Returns:
            None.
        """
        if self.type in ['continuous', 'hcategorical']:
            # random imputation
            if 'random' in imputation_method:
                if len(imputation_method) > 6:
                    try:
                        seed = int(imputation_method[7:])
                        random.seed(seed)
                    except Exception as e:
                        print(f"Warning: Seed must be an integer. Random seed will be set to None. str({e})")
                        random.seed(a=None)
                else:
                    random.seed(a=None)
                self.df_features[var_of_type] = self.df_features[var_of_type].apply(lambda x: x.fillna(random.choice(list(x.dropna(axis=0)))))
            
            # Imputation with median
            elif 'median' in imputation_method:
                self.df_features[var_of_type] = self.df_features[var_of_type].fillna(self.df_features[var_of_type].median())
            
            # Imputation with mean
            elif 'mean' in imputation_method:
                self.df_features[var_of_type] = self.df_features[var_of_type].fillna(self.df_features[var_of_type].mean())
            
            else:
                raise ValueError("Imputation method for continuous and hcategorical features must be 'random', 'median' or 'mean'.")
        
        elif self.type in ['icategorical']:
            if 'random' in imputation_method:
                if len(imputation_method) > 6:
                    seed = int(imputation_method[7:])
                    random.seed(seed)
                else:
                    random.seed(a=None)

                self.df_features[var_of_type] = self.df_features[var_of_type].apply(lambda x: x.fillna(random.choice(list(x.dropna(axis=0)))))

            if 'mode' in imputation_method:
                self.df_features[var_of_type] = self.df_features[var_of_type].fillna(self.df_features[var_of_type].mode().max())
        else:
            raise ValueError("Variable type must be 'continuous', 'hcategorical' or 'icategorical'.")
        
    def __call__(self, cleaning_dict: Dict, imputation_method: str = "mean", 
                missing_cutoff_ps: float = 0.25, missing_cutoff_pf: float = 0.1, 
                cov_cutoff:float = 0.1) -> pd.DataFrame:
        """
        Applies data cleaning to the features of type.

        Args:
            cleaning_dict (dict): Dictionary of cleaning parameters (missing cutoffs and coefficient of variation cutoffs etc.).
            var_of_type (list, optional): List of variable names.
            imputation_method (str): Method of imputation. Can be "mean", "median", "mode" or "random".
                For "random" imputation, a seed can be provided by adding the seed value after the method 
                name, for example "random42".
            missing_cutoff_ps (float, optional): maximal percentage cut-offs of missing features per sample.
            missing_cutoff_pf (float, optional): maximal percentage cut-offs of missing samples per variable.
            cov_cutoff (float, optional): minimal coefficient of variation cut-offs over samples per variable.
        
        Returns:
            pd.DataFrame: Cleaned table of features.
        """

        # Initialization
        var_of_type = self.df_features.Properties['userData']['variables']['continuous']

        # Retrieve thresholds from cleaning_dict if not None
        if cleaning_dict is not None:
            missing_cutoff_pf = cleaning_dict['missingCutoffpf']
            missing_cutoff_ps = cleaning_dict['missingCutoffps']
            cov_cutoff = cleaning_dict['covCutoff']
            imputation_method = cleaning_dict['imputation']
        
        # Replace infinite values with NaNs
        self.df_features = self.df_features.replace([np.inf, -np.inf], np.nan)

        # Remove features with more than missing_cutoff_pf missing samples (NaNs)
        var_of_type = self.cut_off_missing_per_feature(var_of_type, missing_cutoff_pf)

        # Check
        if len(var_of_type) == 0:
            return None

        # Remove features with a coefficient of variation less than cov_cutoff
        var_of_type = self.cut_off_variation(var_of_type, cov_cutoff)

        # Check
        if len(var_of_type) == 0:
            return None

        # Remove scans with more than missing_cutoff_ps missing features
        self.cut_off_missing_per_sample(var_of_type, missing_cutoff_ps)

        # Impute missing values
        self.impute_missing(var_of_type, imputation_method)

        return self.df_features



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np


def batch_patients(n_patient: int,
                   n_batch: int) -> np.ndarray:
    """Replaces volume intensities outside the ROI with NaN.

    Args:
        n_patient (int): Number of patient.
        n_batch (int): Number of batch, usually less or equal to the cores number on your machine.

    Returns:
        ndarray: List of indexes with size n_batch and max value n_patient.
    """

    # FIND THE NUMBER OF PATIENTS IN EACH BATCH
    patients = [0] * n_batch  # np.zeros(n_batch, dtype=int)
    patient_vect = np.random.permutation(n_patient)  # To randomize stuff a bit.
    if n_batch:
        n_p = n_patient / n_batch
        n_sup = np.ceil(n_p).astype(int)
        n_inf = np.floor(n_p).astype(int)
        if n_sup != n_inf:
            n_sub_inf = n_batch - 1
            n_sub_sup = 1
            total = n_sub_inf*n_inf + n_sub_sup*n_sup
            while total != n_patient:
                n_sub_inf = n_sub_inf - 1
                n_sub_sup = n_sub_sup + 1
                total = n_sub_inf*n_inf + n_sub_sup*n_sup

            n_p = np.hstack((np.tile(n_inf, (1, n_sub_inf))[
                           0], np.tile(n_sup, (1, n_sub_sup))[0]))
        else:  # The number of patients in all batches will be the same
            n_p = np.tile(n_sup, (1, n_batch))[0]

        start = 0
        for i in range(0, n_batch):
            patients[i] = patient_vect[start:(start+n_p[i])].tolist()
            start += n_p[i]

    return patients



from string import Template

glcm_kernel = Template("""
#include <stdio.h>
#include <math.h>
#include <iostream>

# define MAX_SIZE ${max_vol}
# define FILTER_SIZE ${filter_size}

// Function flatten a 3D matrix into a 1D vector
__device__ float * reshape(float(*matrix)[FILTER_SIZE][FILTER_SIZE]) {
    //size of array
    const int size = FILTER_SIZE* FILTER_SIZE* FILTER_SIZE;
    float flattened[size];
    int index = 0;
    for (int i = 0; i < FILTER_SIZE; ++i) {
		for (int j = 0; j < FILTER_SIZE; ++j) {
			for (int k = 0; k < FILTER_SIZE; ++k) {
				flattened[index] = matrix[i][j][k];
                index++;
			}
		}
	}
    return flattened;
}

// Function to perform histogram equalisation of the ROI imaging intensities
__device__ void discretize(float vol_quant_re[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE], 
                            float max_val, float min_val=${min_val}) {
    
    // PARSING ARGUMENTS
    float n_q = ${n_q};
    const char* discr_type = "${discr_type}";

    // DISCRETISATION
    if (discr_type == "FBS") {
        float w_b = n_q;
        for (int i = 0; i < FILTER_SIZE; i++) {
            for (int j = 0; j < FILTER_SIZE; j++) {
                for (int k = 0; k < FILTER_SIZE; k++) {
                    float value = vol_quant_re[i][j][k];
                    if (!isnan(value)) {
                        vol_quant_re[i][j][k] = floorf((value - min_val) / w_b) + 1.0;
                    }
                }
            }
        }
    }
    else if (discr_type == "FBN") {
        float w_b = (max_val - min_val) / n_q;
        for (int i = 0; i < FILTER_SIZE; i++) {
            for (int j = 0; j < FILTER_SIZE; j++) {
                for (int k = 0; k < FILTER_SIZE; k++) {
                    float value = vol_quant_re[i][j][k];
                    if (!isnan(value)) {
                        vol_quant_re[i][j][k] = floorf(n_q * ((value - min_val) / (max_val - min_val))) + 1.0;
                        if (value == max_val) {
							vol_quant_re[i][j][k] = n_q;
						}
                    }
                }
            }
        }
    }
    else {
        printf("ERROR: discretization type not supported");
        assert(false);
    }
}

// Compute the diagonal probability
__device__ float * GLCMDiagProb(float p_ij[MAX_SIZE][MAX_SIZE], float max_vol) {
    int valK[MAX_SIZE];
    for (int i = 0; i < (int)max_vol; ++i) {
        valK[i] = i;
    }
    float p_iminusj[MAX_SIZE] = { 0.0 };
    for (int iterationK = 0; iterationK < (int)max_vol; ++iterationK) {
        int k = valK[iterationK];
        float p = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                if (k - fabsf(i - j) == 0) {
                    p += p_ij[i][j];
                }
            }
        }

        p_iminusj[iterationK] = p;
    }

    return p_iminusj;
}

// Compute the cross-diagonal probability
__device__ float * GLCMCrossDiagProb(float p_ij[MAX_SIZE][MAX_SIZE], float max_vol) {
    float valK[2 * MAX_SIZE - 1];
    // fill valK with 2, 3, 4, ..., 2*max_vol - 1
    for (int i = 0; i < 2 * (int)max_vol - 1; ++i) {
        valK[i] = i + 2;
    }
    float p_iplusj[2*MAX_SIZE - 1] = { 0.0 };

    for (int iterationK = 0; iterationK < 2*(int)max_vol - 1; ++iterationK) {
        int k = valK[iterationK];
        float p = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                if (k - (i + j + 2) == 0) {
                    p += p_ij[i][j];
                }
            }
        }

        p_iplusj[iterationK] = p;
    }

    return p_iplusj;
}

__device__ void getGLCMmatrix(
    float (*ROIonly)[FILTER_SIZE][FILTER_SIZE], 
    float GLCMfinal[MAX_SIZE][MAX_SIZE], 
    float max_vol, 
    bool distCorrection = true) 
{
    // PARSING "distCorrection" ARGUMENT
   
    const int Ng = MAX_SIZE;
    float levels[Ng] = {0};
    // initialize levels to 1, 2, 3, ..., 15
    for (int i = 0; i < (int)max_vol; ++i) {
		levels[i] = i + 1;
	}

    float levelTemp = max_vol + 1;

    for (int i = 0; i < FILTER_SIZE; ++i) {
        for (int j = 0; j < FILTER_SIZE; ++j) {
            for (int k = 0; k < FILTER_SIZE; ++k) {
                if (isnan(ROIonly[i][j][k])) {
                    ROIonly[i][j][k] = levelTemp;
                }
            }
        }
    }

    int dim_x = FILTER_SIZE;
    int dim_y = FILTER_SIZE;
    int dim_z = FILTER_SIZE;

    // Reshape the 3D matrix to a 1D vector
    float *q2;
    q2 = reshape(ROIonly);

    // Combine levels and level_temp into qs
    float qs[Ng + 1] = {0};
    for (int i = 0; i < (int)max_vol + 1; ++i) {
        if (i == (int)max_vol) {
			qs[i] = levelTemp;
            break;
		}
        qs[i] = levels[i];
	}
    const int lqs = Ng + 1;

    // Create a q3 matrix and assign values based on qs
    int q3[FILTER_SIZE* FILTER_SIZE* FILTER_SIZE] = {0};

    // fill q3 with 0s
    for (int i = 0; i < FILTER_SIZE * FILTER_SIZE * FILTER_SIZE; ++i) {
        q3[i] = 0;
    }
    for (int k = 0; k < (int)max_vol + 1; ++k) {
        for (int i = 0; i < FILTER_SIZE * FILTER_SIZE * FILTER_SIZE; ++i) {
            if (fabsf(q2[i] - qs[k]) < 1.19209e-07) {
                q3[i] = k;
            }
        }
    }

    // Reshape q3 back to the original dimensions (dimX, dimY, dimZ)
    float reshaped_q3[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE];

    int index = 0;
    for (int i = 0; i < dim_x; ++i) {
        for (int j = 0; j < dim_y; ++j) {
            for (int k = 0; k < dim_z; ++k) {
                reshaped_q3[i][j][k] = q3[index++];
            }
        }
    }


    float GLCM[lqs][lqs] = {0};

    // fill GLCM with 0s
    for (int i = 0; i < (int)max_vol + 1; ++i) {
        for (int j = 0; j < (int)max_vol + 1; ++j) {
			GLCM[i][j] = 0;
		}
	}

    for (int i = 1; i <= dim_x; ++i) {
        int i_min = max(1, i - 1);
        int i_max = min(i + 1, dim_x);
        for (int j = 1; j <= dim_y; ++j) {
            int j_min = max(1, j - 1);
            int j_max = min(j + 1, dim_y);
            for (int k = 1; k <= dim_z; ++k) {
                int k_min = max(1, k - 1);
                int k_max = min(k + 1, dim_z);
                int val_q3 = reshaped_q3[i - 1][j - 1][k - 1];
                for (int I2 = i_min; I2 <= i_max; ++I2) {
                    for (int J2 = j_min; J2 <= j_max; ++J2) {
                        for (int K2 = k_min; K2 <= k_max; ++K2) {
                            if (I2 == i && J2 == j && K2 == k) {
                                continue;
                            }
                            else {
                                int val_neighbor = reshaped_q3[I2 - 1][J2 - 1][K2 - 1];
                                if (distCorrection) {
                                    // Discretization length correction
                                    GLCM[val_q3][val_neighbor] +=
                                        sqrtf(fabsf(I2 - i) +
                                            fabsf(J2 - j) +
                                            fabsf(K2 - k));
                                }
                                else {
                                    GLCM[val_q3][val_neighbor] += 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    // Eliminate last row and column
    for (int i = 0; i < (int)max_vol; ++i) {
		for (int j = 0; j < (int)max_vol; ++j) {
			GLCMfinal[i][j] = GLCM[i][j];
		}
	}
}

__device__ void computeGLCMFeatures(float(*vol)[FILTER_SIZE][FILTER_SIZE], float features[25], float max_vol, bool distCorrection) {

    float GLCM[MAX_SIZE][MAX_SIZE] = { 0.0 };

    // Call function with specified distCorrection
    getGLCMmatrix(vol, GLCM, max_vol, distCorrection);

    // Normalize GLCM
    float sumGLCM = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            sumGLCM += GLCM[i][j];
        }
    }
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            GLCM[i][j] /= sumGLCM;
        }
    }

    // Compute textures
    // // Number of gray levels
    const int Ng = MAX_SIZE;
    float vectNg[Ng];

    // fill vectNg with 1, 2, ..., Ng
    for (int i = 0; i < (int)max_vol; ++i) {
        vectNg[i] = i + 1;
    }

    // Create meshgird of size Ng x Ng
    float colGrid[Ng][Ng] = { 0.0 };
    float rowGrid[Ng][Ng] = { 0.0 };

    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            colGrid[i][j] = vectNg[j];
        }
    }
    for (int j = 0; j < int(max_vol); ++j) {
        for (int i = 0; i < int(max_vol); ++i) {
            rowGrid[i][j] = vectNg[i];
        }
    }
    int step_i = 0;
    int step_j = 0;

    // Joint maximum
    float joint_max = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            joint_max = max(joint_max, GLCM[i][j]);
        }
    }
    features[0] = joint_max;

    // Joint average
    float u = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        step_i = 0;
        for (int j = 0; j < (int)max_vol; ++j) {
            u += GLCM[i][j] * rowGrid[i][j];
            step_i++;
        }
        step_j++;
    }
    features[1] = u;

    // Joint variance
    step_j = 0;
    float var = 0.0;
    u = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        step_i = 0;
        for (int j = 0; j < (int)max_vol; ++j) {
            u += GLCM[i][j] * rowGrid[i][j];
            step_i++;
        }
        step_j++;
    }
    for (int i = 0; i < (int)max_vol; ++i) {
        step_i = 0;
        for (int j = 0; j < (int)max_vol; ++j) {
            var += GLCM[i][j] * powf(rowGrid[i][j] - u, 2);
            step_i++;
        }
        step_j++;
    }
    features[2] = var;

    // Joint entropy
    float entropy = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            if (GLCM[i][j] > 0.0) {
                entropy += GLCM[i][j] * log2f(GLCM[i][j]);
            }
        }
    }
    features[3] = -entropy;

    // Difference average
    float* p_iminusj;
    p_iminusj = GLCMDiagProb(GLCM, max_vol);
    float diff_avg = 0.0;
    float k[Ng];
    // fill k with 0, 1, ..., Ng - 1
    for (int i = 0; i < int(max_vol); ++i) {
        k[i] = i;
    }
    for (int i = 0; i < int(max_vol); ++i) {
        diff_avg += p_iminusj[i] * k[i];
    }
    features[4] = diff_avg;

    // Difference variance
    diff_avg = 0.0;
    // fill k with 0, 1, ..., Ng - 1
    for (int i = 0; i < int(max_vol); ++i) {
        k[i] = i;
    }
    for (int i = 0; i < int(max_vol); ++i) {
        diff_avg += p_iminusj[i] * k[i];
    }
    float diff_var = 0.0;
    step_i = 0;
    for (int i = 0; i < int(max_vol); ++i) {
        diff_var += p_iminusj[i] * powf(k[i] - diff_avg, 2);
        step_i++;
    }
    features[5] = diff_var;

    // Difference entropy
    float diff_entropy = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        if (p_iminusj[i] > 0.0) {
            diff_entropy += p_iminusj[i] * log2f(p_iminusj[i]);
        }
    }
    features[6] = -diff_entropy;

    // Sum average
    float k1[2 * Ng - 1];
    // fill k with 2, 3, ..., 2 * Ng
    for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
        k1[i] = i + 2;
    }
    float sum_avg = 0.0;
    float* p_iplusj = GLCMCrossDiagProb(GLCM, max_vol);
    for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
        sum_avg += p_iplusj[i] * k1[i];
    }
    features[7] = sum_avg;

    // Sum variance
    float sum_var = 0.0;
    for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
        sum_var += p_iplusj[i] * powf(k1[i] - sum_avg, 2);
    }
    features[8] = sum_var;

    // Sum entropy
    float sum_entr = 0.0;
    for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
        if (p_iplusj[i] > 0.0) {
            sum_entr += p_iplusj[i] * log2f(p_iplusj[i]);
        }
    }
    features[9] = -sum_entr;

    // Angular second moment (energy)
    float energy = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            energy += powf(GLCM[i][j], 2);
        }
    }
    features[10] = energy;

    // Contrast
    float contrast = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            contrast += powf(rowGrid[i][j] - colGrid[i][j], 2) * GLCM[i][j];
        }
    }
    features[11] = contrast;

    // Dissimilarity
    float dissimilarity = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            dissimilarity += fabsf(rowGrid[i][j] - colGrid[i][j]) * GLCM[i][j];
        }
    }
    features[12] = dissimilarity;

    // Inverse difference
    float inv_diff = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            inv_diff += GLCM[i][j] / (1 + fabsf(rowGrid[i][j] - colGrid[i][j]));
        }
    }
    features[13] = inv_diff;

    // Inverse difference normalized
    float invDiffNorm = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            invDiffNorm += GLCM[i][j] / (1 + fabsf(rowGrid[i][j] - colGrid[i][j]) / int(max_vol));
        }
    }
    features[14] = invDiffNorm;

    // Inverse difference moment
    float invDiffMom = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            invDiffMom += GLCM[i][j] / (1 + powf((rowGrid[i][j] - colGrid[i][j]), 2));
        }
    }
    features[15] = invDiffMom;

    // Inverse difference moment normalized
    float invDiffMomNorm = 0.0;
    for (int i = 0; i < int(max_vol); ++i) {
        for (int j = 0; j < int(max_vol); ++j) {
            invDiffMomNorm += GLCM[i][j] / (1 + powf((rowGrid[i][j] - colGrid[i][j]), 2) / powf(int(max_vol), 2));
        }
    }
    features[16] = invDiffMomNorm;

    // Inverse variance
    float invVar = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = i + 1; j < int(max_vol); j++) {
            invVar += GLCM[i][j] / powf((i - j), 2);
        }
    }
    features[17] = 2*invVar;

    // Correlation
    float u_i = 0.0;
    float u_j = 0.0;
    float std_i = 0.0;
    float std_j = 0.0;
    float p_i[Ng] = { 0.0 };
    float p_j[Ng] = { 0.0 };

    // sum over rows
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            p_i[i] += GLCM[i][j];
        }
    }
    // sum over columns
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            p_j[j] += GLCM[i][j];
        }
    }
    for (int i = 0; i < int(max_vol); i++) {
        u_i += vectNg[i] * p_i[i];
        u_j += vectNg[i] * p_j[i];
    }
    for (int i = 0; i < int(max_vol); i++) {
        std_i += powf(vectNg[i] - u_i, 2) * p_i[i];
        std_j += powf(vectNg[i] - u_j, 2) * p_j[i];
    }
    std_i = sqrtf(std_i);
    std_j = sqrtf(std_j);

    float tempSum = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            tempSum += rowGrid[i][j] * colGrid[i][j] * GLCM[i][j];
        }
    }
    float correlation = (1 / (std_i * std_j)) * (-u_i * u_j + tempSum);
    features[18] = correlation;

    // Autocorrelation
    float autoCorr = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            autoCorr += rowGrid[i][j] * colGrid[i][j] * GLCM[i][j];
        }
    }
    features[19] = autoCorr;

    // Cluster tendency
    float clusterTend = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            clusterTend += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 2) * GLCM[i][j];
        }
    }
    features[20] = clusterTend;

    // Cluster shade
    float clusterShade = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            clusterShade += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 3) * GLCM[i][j];
        }
    }
    features[21] = clusterShade;

    // Cluster prominence
    float clusterProm = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            clusterProm += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 4) * GLCM[i][j];
        }
    }
    features[22] = clusterProm;

    // First measure of information correlation
    float HXY = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            if (GLCM[i][j] > 0.0) {
                HXY += GLCM[i][j] * log2f(GLCM[i][j]);
            }
        }
    }
    HXY = -HXY;

    float HX = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        if (p_i[i] > 0.0) {
            HX += p_i[i] * log2f(p_i[i]);
        }
    }
    HX = -HX;

    // Repeat p_i and p_j Ng times
    float p_i_temp[Ng][Ng];
    float p_j_temp[Ng][Ng];
    float p_temp[Ng][Ng];

    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            p_i_temp[i][j] = p_i[i];
            p_j_temp[i][j] = p_j[j];
            p_temp[i][j] = p_i_temp[i][j] * p_j_temp[i][j];
        }
    }

    float HXY1 = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            if (p_temp[i][j] > 0.0) {
                HXY1 += GLCM[i][j] * log2f(p_temp[i][j]);
            }
        }
    }
    HXY1 = -HXY1;
    features[23] = (HXY - HXY1) / HX;

    // Second measure of information correlation
    float HXY2 = 0.0;
    for (int i = 0; i < int(max_vol); i++) {
        for (int j = 0; j < int(max_vol); j++) {
            if (p_temp[i][j] > 0.0) {
                HXY2 += p_temp[i][j] * log2f(p_temp[i][j]);
            }
        }
    }
    HXY2 = -HXY2;
    if (HXY > HXY2) {
        features[24] = 0.0;
    }
    else {
        features[24] = sqrtf(1 - expf(-2 * (HXY2 - HXY)));
    }
}

extern "C"
__global__ void glcm_filter_global(
    float vol[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}][25],
    float vol_copy[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    bool distCorrection = false)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    int j = blockIdx.y * blockDim.y + threadIdx.y;
    int k = blockIdx.z * blockDim.z + threadIdx.z;

    if (i < ${shape_volume_0} && j < ${shape_volume_1} && k < ${shape_volume_2} && i >= 0 && j >= 0 && k >= 0) {
        // pad size
        const int padd_size = (FILTER_SIZE - 1) / 2;

        // size vol
        const int size_x = ${shape_volume_0};
        const int size_y = ${shape_volume_1};
        const int size_z = ${shape_volume_2};

        // skip all calculations if vol at position i,j,k is nan
        if (!isnan(vol_copy[i][j][k])) {
            // get submatrix
            float sub_matrix[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE] = {NAN};
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                    if ((i - padd_size + idx_i) >= 0 && (i - padd_size + idx_i) < size_x && 
                        (j - padd_size + idx_j) >= 0 && (j - padd_size + idx_j) < size_y &&
                        (k - padd_size + idx_k) >= 0 && (k - padd_size + idx_k) < size_z) {
                            sub_matrix[idx_i][idx_j][idx_k] = vol_copy[i - padd_size + idx_i][j - padd_size + idx_j][k - padd_size + idx_k];
                        }
                    }
                }
            }

            // get the maximum value of the submatrix
            float max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }
            
            // compute GLCM features
            float features[25] = { 0.0 };
            computeGLCMFeatures(sub_matrix, features, max_vol, false);
            
            // Copy GLCM feature to voxels of the volume
            if (i < size_x && j < size_y && k < size_z){
                for (int idx = 0; idx < 25; ++idx) {
                    vol[i][j][k][idx] = features[idx];
                }
            }
        }
    }
}

extern "C"
__global__ void glcm_filter_local(
    float vol[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}][25],
    float vol_copy[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    bool distCorrection = false)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    int j = blockIdx.y * blockDim.y + threadIdx.y;
    int k = blockIdx.z * blockDim.z + threadIdx.z;

    if (i < ${shape_volume_0} && j < ${shape_volume_1} && k < ${shape_volume_2} && i >= 0 && j >= 0 && k >= 0) {
        // pad size
        const int padd_size = (FILTER_SIZE - 1) / 2;

        // size vol
        const int size_x = ${shape_volume_0};
        const int size_y = ${shape_volume_1};
        const int size_z = ${shape_volume_2};

        // skip all calculations if vol at position i,j,k is nan
        if (!isnan(vol_copy[i][j][k])) {
            // get submatrix
            float sub_matrix[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE] = {NAN};
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                    if ((i - padd_size + idx_i) >= 0 && (i - padd_size + idx_i) < size_x && 
                        (j - padd_size + idx_j) >= 0 && (j - padd_size + idx_j) < size_y &&
                        (k - padd_size + idx_k) >= 0 && (k - padd_size + idx_k) < size_z) {
                            sub_matrix[idx_i][idx_j][idx_k] = vol_copy[i - padd_size + idx_i][j - padd_size + idx_j][k - padd_size + idx_k];
                        }
                    }
                }
            }

            // get the maximum value of the submatrix
            float max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }
            // get the minimum value of the submatrix if discr_type is FBN
            float min_val = 3.40282e+38;
            if ("${discr_type}" == "FBN") {
                for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                    for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                        for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                            min_val = min(min_val, sub_matrix[idx_i][idx_j][idx_k]);
                        }
                    }
                }
                discretize(sub_matrix, max_vol, min_val);
            }

            // If FBS discretize the submatrix with user set minimum value
            else{
                discretize(sub_matrix, max_vol);
            }
                                                          
            // get the maximum value of the submatrix after discretization
            max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }

            // compute GLCM features
            float features[25] = { 0.0 };
            computeGLCMFeatures(sub_matrix, features, max_vol, false);
            
            // Copy GLCM feature to voxels of the volume
            if (i < size_x && j < size_y && k < size_z){
                for (int idx = 0; idx < 25; ++idx) {
                    vol[i][j][k][idx] = features[idx];
                }
            }
        }
    }
}
""")
                       
# Signle-feature kernel
single_glcm_kernel = Template("""
#include <stdio.h>
#include <math.h>
#include <iostream>

# define MAX_SIZE ${max_vol}
# define FILTER_SIZE ${filter_size}

// Function flatten a 3D matrix into a 1D vector
__device__ float * reshape(float(*matrix)[FILTER_SIZE][FILTER_SIZE]) {
    //size of array
    const int size = FILTER_SIZE* FILTER_SIZE* FILTER_SIZE;
    float flattened[size];
    int index = 0;
    for (int i = 0; i < FILTER_SIZE; ++i) {
		for (int j = 0; j < FILTER_SIZE; ++j) {
			for (int k = 0; k < FILTER_SIZE; ++k) {
				flattened[index] = matrix[i][j][k];
                index++;
			}
		}
	}
    return flattened;
}

// Function to perform discretization on the ROI imaging intensities
__device__ void discretize(float vol_quant_re[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE], 
                            float max_val, float min_val=${min_val}) {
    
    // PARSING ARGUMENTS
    float n_q = ${n_q};
    const char* discr_type = "${discr_type}";

    // DISCRETISATION
    if (discr_type == "FBS") {
        float w_b = n_q;
        for (int i = 0; i < FILTER_SIZE; i++) {
            for (int j = 0; j < FILTER_SIZE; j++) {
                for (int k = 0; k < FILTER_SIZE; k++) {
                    float value = vol_quant_re[i][j][k];
                    if (!isnan(value)) {
                        vol_quant_re[i][j][k] = floorf((value - min_val) / w_b) + 1.0;
                    }
                }
            }
        }
    }
    else if (discr_type == "FBN") {
        float w_b = (max_val - min_val) / n_q;
        for (int i = 0; i < FILTER_SIZE; i++) {
            for (int j = 0; j < FILTER_SIZE; j++) {
                for (int k = 0; k < FILTER_SIZE; k++) {
                    float value = vol_quant_re[i][j][k];
                    if (!isnan(value)) {
                        vol_quant_re[i][j][k] = floorf(n_q * ((value - min_val) / (max_val - min_val))) + 1.0;
                        if (value == max_val) {
							vol_quant_re[i][j][k] = n_q;
						}
                    }
                }
            }
        }
    }
    else {
        printf("ERROR: discretization type not supported");
        assert(false);
    }
}

// Compute the diagonal probability
__device__ float * GLCMDiagProb(float p_ij[MAX_SIZE][MAX_SIZE], float max_vol) {
    int valK[MAX_SIZE];
    for (int i = 0; i < (int)max_vol; ++i) {
        valK[i] = i;
    }
    float p_iminusj[MAX_SIZE] = { 0.0 };
    for (int iterationK = 0; iterationK < (int)max_vol; ++iterationK) {
        int k = valK[iterationK];
        float p = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                if (k - fabsf(i - j) == 0) {
                    p += p_ij[i][j];
                }
            }
        }

        p_iminusj[iterationK] = p;
    }

    return p_iminusj;
}

// Compute the cross-diagonal probability
__device__ float * GLCMCrossDiagProb(float p_ij[MAX_SIZE][MAX_SIZE], float max_vol) {
    float valK[2 * MAX_SIZE - 1];
    // fill valK with 2, 3, 4, ..., 2*max_vol - 1
    for (int i = 0; i < 2 * (int)max_vol - 1; ++i) {
        valK[i] = i + 2;
    }
    float p_iplusj[2*MAX_SIZE - 1] = { 0.0 };

    for (int iterationK = 0; iterationK < 2*(int)max_vol - 1; ++iterationK) {
        int k = valK[iterationK];
        float p = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                if (k - (i + j + 2) == 0) {
                    p += p_ij[i][j];
                }
            }
        }

        p_iplusj[iterationK] = p;
    }

    return p_iplusj;
}

__device__ void getGLCMmatrix(
    float (*ROIonly)[FILTER_SIZE][FILTER_SIZE], 
    float GLCMfinal[MAX_SIZE][MAX_SIZE], 
    float max_vol, 
    bool distCorrection = true) 
{
    // PARSING "distCorrection" ARGUMENT
   
    const int Ng = MAX_SIZE;
    float levels[Ng] = {0};
    // initialize levels to 1, 2, 3, ..., 15
    for (int i = 0; i < (int)max_vol; ++i) {
		levels[i] = i + 1;
	}

    float levelTemp = max_vol + 1;

    for (int i = 0; i < FILTER_SIZE; ++i) {
        for (int j = 0; j < FILTER_SIZE; ++j) {
            for (int k = 0; k < FILTER_SIZE; ++k) {
                if (isnan(ROIonly[i][j][k])) {
                    ROIonly[i][j][k] = levelTemp;
                }
            }
        }
    }

    int dim_x = FILTER_SIZE;
    int dim_y = FILTER_SIZE;
    int dim_z = FILTER_SIZE;

    // Reshape the 3D matrix to a 1D vector
    float *q2;
    q2 = reshape(ROIonly);

    // Combine levels and level_temp into qs
    float qs[Ng + 1] = {0};
    for (int i = 0; i < (int)max_vol + 1; ++i) {
        if (i == (int)max_vol) {
			qs[i] = levelTemp;
            break;
		}
        qs[i] = levels[i];
	}
    const int lqs = Ng + 1;

    // Create a q3 matrix and assign values based on qs
    int q3[FILTER_SIZE* FILTER_SIZE* FILTER_SIZE] = {0};

    // fill q3 with 0s
    for (int i = 0; i < FILTER_SIZE * FILTER_SIZE * FILTER_SIZE; ++i) {
        q3[i] = 0;
    }
    for (int k = 0; k < (int)max_vol + 1; ++k) {
        for (int i = 0; i < FILTER_SIZE * FILTER_SIZE * FILTER_SIZE; ++i) {
            if (fabsf(q2[i] - qs[k]) < 1.19209e-07) {
                q3[i] = k;
            }
        }
    }

    // Reshape q3 back to the original dimensions (dimX, dimY, dimZ)
    float reshaped_q3[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE];

    int index = 0;
    for (int i = 0; i < dim_x; ++i) {
        for (int j = 0; j < dim_y; ++j) {
            for (int k = 0; k < dim_z; ++k) {
                reshaped_q3[i][j][k] = q3[index++];
            }
        }
    }


    float GLCM[lqs][lqs] = {0};

    // fill GLCM with 0s
    for (int i = 0; i < (int)max_vol + 1; ++i) {
        for (int j = 0; j < (int)max_vol + 1; ++j) {
			GLCM[i][j] = 0;
		}
	}

    for (int i = 1; i <= dim_x; ++i) {
        int i_min = max(1, i - 1);
        int i_max = min(i + 1, dim_x);
        for (int j = 1; j <= dim_y; ++j) {
            int j_min = max(1, j - 1);
            int j_max = min(j + 1, dim_y);
            for (int k = 1; k <= dim_z; ++k) {
                int k_min = max(1, k - 1);
                int k_max = min(k + 1, dim_z);
                int val_q3 = reshaped_q3[i - 1][j - 1][k - 1];
                for (int I2 = i_min; I2 <= i_max; ++I2) {
                    for (int J2 = j_min; J2 <= j_max; ++J2) {
                        for (int K2 = k_min; K2 <= k_max; ++K2) {
                            if (I2 == i && J2 == j && K2 == k) {
                                continue;
                            }
                            else {
                                int val_neighbor = reshaped_q3[I2 - 1][J2 - 1][K2 - 1];
                                if (distCorrection) {
                                    // Discretization length correction
                                    GLCM[val_q3][val_neighbor] +=
                                        sqrtf(fabsf(I2 - i) +
                                            fabsf(J2 - j) +
                                            fabsf(K2 - k));
                                }
                                else {
                                    GLCM[val_q3][val_neighbor] += 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    // Eliminate last row and column
    for (int i = 0; i < (int)max_vol; ++i) {
		for (int j = 0; j < (int)max_vol; ++j) {
			GLCMfinal[i][j] = GLCM[i][j];
		}
	}
}

__device__ float computeGLCMFeatures(float(*vol)[FILTER_SIZE][FILTER_SIZE], int feature, float max_vol, bool distCorrection) {

    float GLCM[MAX_SIZE][MAX_SIZE] = { 0.0 };

    // Call function with specified distCorrection
    getGLCMmatrix(vol, GLCM, max_vol, distCorrection);

    // Normalize GLCM
    float sumGLCM = 0.0;
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
			sumGLCM += GLCM[i][j];
		}
	}
    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            GLCM[i][j] /= sumGLCM;
        }
    }

    // Compute textures
    // // Number of gray levels
    const int Ng = MAX_SIZE;
    float vectNg[Ng];

    // fill vectNg with 1, 2, ..., Ng
    for (int i = 0; i < (int)max_vol; ++i) {
		vectNg[i] = i + 1;
	}

    // Create meshgird of size Ng x Ng
    float colGrid[Ng][Ng] = { 0.0 };
    float rowGrid[Ng][Ng] = { 0.0 };

    for (int i = 0; i < (int)max_vol; ++i) {
        for (int j = 0; j < (int)max_vol; ++j) {
            colGrid[i][j] = vectNg[j];
        }
    }
    for (int j = 0; j < int(max_vol); ++j) {
        for (int i = 0; i < int(max_vol); ++i) {
            rowGrid[i][j] = vectNg[i];
        }
    }
    int step_i = 0;
    int step_j = 0;

    // Joint maximum
    if (feature == 0) {
        float joint_max = NAN; 
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                joint_max = max(joint_max, GLCM[i][j]);
            }
        }
        return joint_max;
    }
    // Joint average
    else if (feature == 1) {
        float u = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            step_i = 0;
            for (int j = 0; j < (int)max_vol; ++j) {
                u += GLCM[i][j] * rowGrid[i][j];
                step_i++;
            }
            step_j++;
        }
        return u;
    }
    // Joint variance
    else if (feature == 2) {
        step_j = 0;
        float var = 0.0;
        float u = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            step_i = 0;
            for (int j = 0; j < (int)max_vol; ++j) {
                u += GLCM[i][j] * rowGrid[i][j];
                step_i++;
            }
            step_j++;
        }
        for (int i = 0; i < (int)max_vol; ++i) {
            step_i = 0;
            for (int j = 0; j < (int)max_vol; ++j) {
                var += GLCM[i][j] * powf(rowGrid[i][j] - u, 2);
                step_i++;
            }
            step_j++;
        }
        return var;
    }
    // Joint entropy
    else if (feature == 3) {
		float entropy = 0.0;
        for (int i = 0; i < (int)max_vol; ++i) {
            for (int j = 0; j < (int)max_vol; ++j) {
                if (GLCM[i][j] > 0.0) {
                    entropy += GLCM[i][j] * log2f(GLCM[i][j]);
				}
			}
		}
		return -entropy;
	}
    // Difference average
    else if (feature == 4) {
        float *p_iminusj;
        p_iminusj = GLCMDiagProb(GLCM, max_vol);
        float diff_avg = 0.0;
        float k[Ng];
        // fill k with 0, 1, ..., Ng - 1
        for (int i = 0; i < int(max_vol); ++i) {
			k[i] = i;
		}
        for (int i = 0; i < int(max_vol); ++i) {
            diff_avg += p_iminusj[i] * k[i];
        }
        return diff_avg;
	}
    // Difference variance
    else if (feature == 5) {
        float* p_iminusj;
        p_iminusj = GLCMDiagProb(GLCM, max_vol);
        float diff_avg = 0.0;
        float k[Ng];
        // fill k with 0, 1, ..., Ng - 1
        for (int i = 0; i < int(max_vol); ++i) {
            k[i] = i;
        }
        for (int i = 0; i < int(max_vol); ++i) {
            diff_avg += p_iminusj[i] * k[i];
        }
        float diff_var = 0.0;
        step_i = 0;
        for (int i = 0; i < int(max_vol); ++i) {
            diff_var += p_iminusj[i] * powf(k[i] - diff_avg, 2);
            step_i++;
        }
        return diff_var;
    }
    // Difference entropy
    else if (feature == 6) {
        float* p_iminusj = GLCMDiagProb(GLCM, max_vol);
		float diff_entropy = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            if (p_iminusj[i] > 0.0) {
				diff_entropy += p_iminusj[i] * log2f(p_iminusj[i]);
			}
		}
		return -diff_entropy;
	}
    // Sum average
    else if (feature == 7) {
        float k[2 * Ng - 1];
        // fill k with 2, 3, ..., 2 * Ng
        for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
			k[i] = i + 2;
		}
        float sum_avg = 0.0;
        float* p_iplusj = GLCMCrossDiagProb(GLCM, max_vol);
        for (int i = 0; i < 2*int(max_vol) - 1; ++i) {
            sum_avg += p_iplusj[i] * k[i];
        }
        return sum_avg;
	}
    // Sum variance
    else if (feature == 8) {
		float k[2 * Ng - 1];
		// fill k with 2, 3, ..., 2 * Ng
        for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
			k[i] = i + 2;
		}
		float sum_avg = 0.0;
		float* p_iplusj = GLCMCrossDiagProb(GLCM, max_vol);
        for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
			sum_avg += p_iplusj[i] * k[i];
		}
		float sum_var = 0.0;
        for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
			sum_var += p_iplusj[i] * powf(k[i] - sum_avg, 2);
		}
		return sum_var;
	}
    // Sum entropy
    else if (feature == 9) {
        float sum_entr = 0.0;
        float* p_iplusj = GLCMCrossDiagProb(GLCM, max_vol);
        for (int i = 0; i < 2 * int(max_vol) - 1; ++i) {
            if (p_iplusj[i] > 0.0) {
                sum_entr += p_iplusj[i] * log2f(p_iplusj[i]);
            }
        }
        return -sum_entr;
    }
    // Angular second moment (energy)
    else if (feature == 10) {
		float energy = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                energy += powf(GLCM[i][j], 2);
			}
		}
		return energy;
	}
    // Contrast
    else if (feature == 11) {
		float contrast = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                contrast += powf(rowGrid[i][j] - colGrid[i][j], 2) * GLCM[i][j];
			}
		}
		return contrast;
	}
    // Dissimilarity
    else if (feature == 12) {
        float dissimilarity = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                dissimilarity += fabsf(rowGrid[i][j] - colGrid[i][j]) * GLCM[i][j];
            }
        }
        return dissimilarity;
    }
    // Inverse difference
    else if (feature == 13) {
		float inv_diff = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                inv_diff += GLCM[i][j] / (1 + fabsf(rowGrid[i][j] - colGrid[i][j]));
			}
		}
		return inv_diff;
	}
    // Inverse difference normalized
    else if (feature == 14) {
        float invDiffNorm = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                invDiffNorm += GLCM[i][j] / (1 + fabsf(rowGrid[i][j] - colGrid[i][j]) / int(max_vol));
            }
        }
        return invDiffNorm;
    }
    // Inverse difference moment
    else if (feature == 15) {
        float invDiffMom = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                invDiffMom += GLCM[i][j] / (1 + powf((rowGrid[i][j] - colGrid[i][j]), 2));
            }
        }
        return invDiffMom;
	}
    // Inverse difference moment normalized
    else if (feature == 16) {
		float invDiffMomNorm = 0.0;
        for (int i = 0; i < int(max_vol); ++i) {
            for (int j = 0; j < int(max_vol); ++j) {
                invDiffMomNorm += GLCM[i][j] / (1 + powf((rowGrid[i][j] - colGrid[i][j]), 2) / powf(int(max_vol), 2));
			}
		}
		return invDiffMomNorm;
	}
    // Inverse variance
    else if (feature == 17) {
        float invVar = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = i + 1; j < int(max_vol); j++) {
                invVar += GLCM[i][j] / powf((i - j), 2);
            }
        }
        return 2*invVar;
    }
    // Correlation
    else if (feature == 18) {
        float u_i = 0.0;
        float u_j = 0.0;
        float std_i = 0.0;
        float std_j = 0.0;
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };

        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }
        for (int i = 0; i < int(max_vol); i++) {
            u_i += vectNg[i] * p_i[i];
            u_j += vectNg[i] * p_j[i];
        }
        for (int i = 0; i < int(max_vol); i++) {
            std_i += powf(vectNg[i] - u_i, 2) * p_i[i];
            std_j += powf(vectNg[i] - u_j, 2) * p_j[i];
        }
        std_i = sqrtf(std_i);
        std_j = sqrtf(std_j);

        float tempSum = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                tempSum += rowGrid[i][j] * colGrid[i][j] * GLCM[i][j];
            }
        }
        float correlation = (1 / (std_i * std_j)) * (-u_i * u_j + tempSum);
        return correlation;
    }
    // Autocorrelation
    else if (feature == 19) {
        float u_i = 0.0;
        float u_j = 0.0;
        float std_i = 0.0;
        float std_j = 0.0;
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };

        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }
        for (int i = 0; i < int(max_vol); i++) {
            u_i += vectNg[i] * p_i[i];
            u_j += vectNg[i] * p_j[i];
        }
        for (int i = 0; i < int(max_vol); i++) {
            std_i += powf(vectNg[i] - u_i, 2) * p_i[i];
            std_j += powf(vectNg[i] - u_j, 2) * p_j[i];
        }
        std_i = sqrtf(std_i);
        std_j = sqrtf(std_j);

        float autoCorr = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                autoCorr += rowGrid[i][j] * colGrid[i][j] * GLCM[i][j];
            }
        }
        
        return autoCorr;
    }
    // Cluster tendency
    else if (feature == 20) {
        float u_i = 0.0;
        float u_j = 0.0;
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };

        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }
        for (int i = 0; i < int(max_vol); i++) {
            u_i += vectNg[i] * p_i[i];
            u_j += vectNg[i] * p_j[i];
        }
        float clusterTend = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                clusterTend += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 2) * GLCM[i][j];
            }
        }
		return clusterTend;
	}
    // Cluster shade
    else if (feature == 21) {
		float u_i = 0.0;
		float u_j = 0.0;
		float p_i[Ng] = { 0.0 };
		float p_j[Ng] = { 0.0 };

		// sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
				p_i[i] += GLCM[i][j];
			}
		}
		// sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
				p_j[j] += GLCM[i][j];
			}
		}
        for (int i = 0; i < int(max_vol); i++) {
			u_i += vectNg[i] * p_i[i];
			u_j += vectNg[i] * p_j[i];
		}
		float clusterShade = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
				clusterShade += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 3) * GLCM[i][j];
			}
		}
		return clusterShade;
	}
    // Cluster prominence
    else if (feature == 22) {
        float u_i = 0.0;
        float u_j = 0.0;
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };

        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }
        for (int i = 0; i < int(max_vol); i++) {
            u_i += vectNg[i] * p_i[i];
            u_j += vectNg[i] * p_j[i];
        }
        float clusterProm = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                clusterProm += powf(rowGrid[i][j] + colGrid[i][j] - u_i - u_j, 4) * GLCM[i][j];
            }
        }
        return clusterProm;
    }
    // First measure of information correlation
    else if (feature == 23) {
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };
        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }

        float HXY = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                if (GLCM[i][j] > 0.0) {
                    HXY += GLCM[i][j] * log2f(GLCM[i][j]);
                }
            }
        }
        HXY = -HXY;

        float HX = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            if (p_i[i] > 0.0) {
                HX += p_i[i] * log2f(p_i[i]);
            }
        }
        HX = -HX;

        // Repeat p_i and p_j Ng times
        float p_i_temp[Ng][Ng];
        float p_j_temp[Ng][Ng];
        float p_temp[Ng][Ng];

        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i_temp[i][j] = p_i[i];
                p_j_temp[i][j] = p_j[j];
                p_temp[i][j] = p_i_temp[i][j] * p_j_temp[i][j];
            }
        }

        float HXY1 = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                if (p_temp[i][j] > 0.0) {
                    HXY1 += GLCM[i][j] * log2f(p_temp[i][j]);
                }
            }
        }
        HXY1 = -HXY1;

        return (HXY - HXY1) / HX;
    }
    // Second measure of information correlation
    else if (feature == 24) {
        float p_i[Ng] = { 0.0 };
        float p_j[Ng] = { 0.0 };
        // sum over rows
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i[i] += GLCM[i][j];
            }
        }
        // sum over columns
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_j[j] += GLCM[i][j];
            }
        }

        float HXY = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                if (GLCM[i][j] > 0.0) {
                    HXY += GLCM[i][j] * log2f(GLCM[i][j]);
                }
            }
        }
        HXY = -HXY;

        // Repeat p_i and p_j Ng times
        float p_i_temp[Ng][Ng];
        float p_j_temp[Ng][Ng];
        float p_temp[Ng][Ng];

        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                p_i_temp[i][j] = p_i[i];
                p_j_temp[i][j] = p_j[j];
                p_temp[i][j] = p_i_temp[i][j] * p_j_temp[i][j];
            }
        }

        float HXY2 = 0.0;
        for (int i = 0; i < int(max_vol); i++) {
            for (int j = 0; j < int(max_vol); j++) {
                if (p_temp[i][j] > 0.0) {
                    HXY2 += p_temp[i][j] * log2f(p_temp[i][j]);
                }
            }
        }
        HXY2 = -HXY2;
        if (HXY > HXY2) {
            return 0.0;
        }
        else {
            return sqrtf(1 - expf(-2 * (HXY2 - HXY)));
        }
    }
    else {
        // Print error message
        printf("Error: feature %d not implemented\\n", feature);
        assert(false);
    }
}

extern "C"
__global__ void glcm_filter_global(
    float vol[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    float vol_copy[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    bool distCorrection = false)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    int j = blockIdx.y * blockDim.y + threadIdx.y;
    int k = blockIdx.z * blockDim.z + threadIdx.z;

    if (i < ${shape_volume_0} && j < ${shape_volume_1} && k < ${shape_volume_2} && i >= 0 && j >= 0 && k >= 0) {
        // pad size
        const int padd_size = (FILTER_SIZE - 1) / 2;

        // size vol
        const int size_x = ${shape_volume_0};
        const int size_y = ${shape_volume_1};
        const int size_z = ${shape_volume_2};

        // skip all calculations if vol at position i,j,k is nan
        if (!isnan(vol_copy[i][j][k])) {
            // get submatrix
            float sub_matrix[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE] = {NAN};
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                    if ((i - padd_size + idx_i) >= 0 && (i - padd_size + idx_i) < size_x && 
                        (j - padd_size + idx_j) >= 0 && (j - padd_size + idx_j) < size_y &&
                        (k - padd_size + idx_k) >= 0 && (k - padd_size + idx_k) < size_z) {
                            sub_matrix[idx_i][idx_j][idx_k] = vol_copy[i - padd_size + idx_i][j - padd_size + idx_j][k - padd_size + idx_k];
                        }
                    }
                }
            }

            // get the maximum value of the submatrix
            float max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }

            // get feature index
            const int feature = ${feature_index};
            
            // compute GLCM features
            float glcm_feature = computeGLCMFeatures(sub_matrix, feature, max_vol, false);
            
            // Copy GLCM feature to voxels of the volume
            if (i < size_x && j < size_y && k < size_z){
                vol[i][j][k] = glcm_feature;
            }
        }
    }
}

extern "C"
__global__ void glcm_filter_local(
    float vol[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    float vol_copy[${shape_volume_0}][${shape_volume_1}][${shape_volume_2}],
    bool distCorrection = false)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    int j = blockIdx.y * blockDim.y + threadIdx.y;
    int k = blockIdx.z * blockDim.z + threadIdx.z;

    if (i < ${shape_volume_0} && j < ${shape_volume_1} && k < ${shape_volume_2} && i >= 0 && j >= 0 && k >= 0) {
        // pad size
        const int padd_size = (FILTER_SIZE - 1) / 2;

        // size vol
        const int size_x = ${shape_volume_0};
        const int size_y = ${shape_volume_1};
        const int size_z = ${shape_volume_2};

        // skip all calculations if vol at position i,j,k is nan
        if (!isnan(vol_copy[i][j][k])) {
            // get submatrix
            float sub_matrix[FILTER_SIZE][FILTER_SIZE][FILTER_SIZE] = {NAN};
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                    if ((i - padd_size + idx_i) >= 0 && (i - padd_size + idx_i) < size_x && 
                        (j - padd_size + idx_j) >= 0 && (j - padd_size + idx_j) < size_y &&
                        (k - padd_size + idx_k) >= 0 && (k - padd_size + idx_k) < size_z) {
                            sub_matrix[idx_i][idx_j][idx_k] = vol_copy[i - padd_size + idx_i][j - padd_size + idx_j][k - padd_size + idx_k];
                        }
                    }
                }
            }

            // get the maximum value of the submatrix
            float max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }
            // get the minimum value of the submatrix if discr_type is FBN
            float min_val = 3.40282e+38;
            if ("${discr_type}" == "FBN") {
                for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                    for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                        for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                            min_val = min(min_val, sub_matrix[idx_i][idx_j][idx_k]);
                        }
                    }
                }
                discretize(sub_matrix, max_vol, min_val);
            }

            // If FBS discretize the submatrix with user set minimum value
            else{
                discretize(sub_matrix, max_vol);
            }
                                                          
            // get the maximum value of the submatrix after discretization
            max_vol = -3.40282e+38;
            for (int idx_i = 0; idx_i < FILTER_SIZE; ++idx_i) {
                for (int idx_j = 0; idx_j < FILTER_SIZE; ++idx_j) {
                    for (int idx_k = 0; idx_k < FILTER_SIZE; ++idx_k) {
                        max_vol = max(max_vol, sub_matrix[idx_i][idx_j][idx_k]);
                    }
                }
            }

            // get feature index
            const int feature = ${feature_index};
            
            // compute GLCM features
            float glcm_feature = computeGLCMFeatures(sub_matrix, feature, max_vol, false);
            
            // Copy GLCM feature to voxels of the volume
            if (i < size_x && j < size_y && k < size_z){
                vol[i][j][k] = glcm_feature;
            }
        }
    }
}
""")


from . import *
from .BatchExtractor import *
from .BatchExtractorTexturalFilters import *
from .diagnostics import *
from .get_oriented_bound_box import *
from .glcm import *
from .gldzm import *
from .glrlm import *
from .glszm import *
from .int_vol_hist import *
from .intensity_histogram import *
from .local_intensity import *
from .morph import *
from .ngldm import *
from .ngtdm import *
from .stats import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Union

import numpy as np


def write_radiomics_csv(path_radiomics_table: Union[Path, str]) -> None:
    """
    Loads a radiomics structure (dict with radiomics features) to convert it to a CSV file and save it.

    Args:
        path_radiomics_table(Union[Path, str]): path to the radiomics dict.
    
    Returns:
        None.
    """

    # INITIALIZATION
    path_radiomics_table = Path(path_radiomics_table)
    path_to_table = path_radiomics_table.parent
    name_table = path_radiomics_table.stem

    # LOAD RADIOMICS TABLE
    radiomics_table_dict = np.load(path_radiomics_table, allow_pickle=True)[0]

    # WRITE RADIOMICS TABLE IN CSV FORMAT
    csv_name = name_table + '.csv'
    csv_path = path_to_table / csv_name
    radiomics_table_dict['Table'] = radiomics_table_dict['Table'].fillna(value='NaN')
    radiomics_table_dict['Table'] = radiomics_table_dict['Table'].sort_index()
    radiomics_table_dict['Table'].to_csv(csv_path,
                                       sep=',',
                                       encoding='utf-8',
                                       index=True,
                                       index_label=radiomics_table_dict['Properties']['DimensionNames'][0])

    # WRITE DEFINITIONS.TXT
    txt_name = name_table + '.txt'
    txt_Path = path_to_table / txt_name

    # WRITE THE CSV
    fid = open(txt_Path, 'w')
    fid.write(radiomics_table_dict['Properties']['UserData'])
    fid.close()



from typing import Dict

import numpy as np

from ..processing.segmentation import compute_bounding_box


def extract_all(vol_obj: np.ndarray,
                      roi_obj_int: np.ndarray,
                      roi_obj_morph: np.ndarray,
                      im_type: str) -> Dict:
    """Computes diagnostic features

    The diagnostic features help identify issues with
    the implementation of the image processing sequence.

    Args:
        vol_obj (ndarray): Imaging data.
        roi_obj_int (ndarray): Intensity mask data.
        roi_obj_morph (ndarray): Morphological mask data.
        im_type (str): Image processing step.

            - 'reSeg': Computes Diagnostic features right after the re-segmentaion step.
            - 'interp' or any other arg: Computes Diagnostic features for any processing step other than re-segmentation.

    Returns:
        Dict: Dictionnary containing the computed features.
    """
    diag = {}

    #  FOR THE IMAGE

    if im_type != 'reSeg':
        # Image dimension x
        diag.update({'image_' + im_type + '_dimX':
                     vol_obj.spatialRef.ImageSize[0]})

        # Image dimension y
        diag.update({'image_' + im_type + '_dimY':
                     vol_obj.spatialRef.ImageSize[1]})

        # Image dimension z
        diag.update({'image_' + im_type + '_dimz':
                     vol_obj.spatialRef.ImageSize[2]})

        # Voxel dimension x
        diag.update({'image_' + im_type + '_voxDimX':
                     vol_obj.spatialRef.PixelExtentInWorldX})

        # Voxel dimension y
        diag.update({'image_' + im_type + '_voxDimY':
                     vol_obj.spatialRef.PixelExtentInWorldY})

        # Voxel dimension z
        diag.update({'image_' + im_type + '_voxDimZ':
                     vol_obj.spatialRef.PixelExtentInWorldZ})

        # Mean intensity
        diag.update({'image_' + im_type + '_meanInt': np.mean(vol_obj.data)})

        # Minimum intensity
        diag.update({'image_' + im_type + '_minInt': np.min(vol_obj.data)})

        # Maximum intensity
        diag.update({'image_' + im_type + '_maxInt': np.max(vol_obj.data)})

    # FOR THE ROI
    box_bound_int = compute_bounding_box(roi_obj_int.data)
    box_bound_morph = compute_bounding_box(roi_obj_morph.data)

    x_gl_int = vol_obj.data[roi_obj_int.data == 1]
    x_gl_morph = vol_obj.data[roi_obj_morph.data == 1]

    # Map dimension x
    diag.update({'roi_' + im_type + '_Int_dimX':
                 roi_obj_int.spatialRef.ImageSize[0]})

    # Map dimension y
    diag.update({'roi_' + im_type + '_Int_dimY':
                 roi_obj_int.spatialRef.ImageSize[1]})

    # Map dimension z
    diag.update({'roi_' + im_type + '_Int_dimZ':
                 roi_obj_int.spatialRef.ImageSize[2]})

    # Bounding box dimension x
    diag.update({'roi_' + im_type + '_Int_boxBoundDimX':
                 box_bound_int[0, 1] - box_bound_int[0, 0] + 1})

    # Bounding box dimension y
    diag.update({'roi_' + im_type + '_Int_boxBoundDimY':
                 box_bound_int[1, 1] - box_bound_int[1, 0] + 1})

    # Bounding box dimension z
    diag.update({'roi_' + im_type + '_Int_boxBoundDimZ':
                 box_bound_int[2, 1] - box_bound_int[2, 0] + 1})

    # Bounding box dimension x
    diag.update({'roi_' + im_type + '_Morph_boxBoundDimX':
                 box_bound_morph[0, 1] - box_bound_morph[0, 0] + 1})

    # Bounding box dimension y
    diag.update({'roi_' + im_type + '_Morph_boxBoundDimY':
                 box_bound_morph[1, 1] - box_bound_morph[1, 0] + 1})

    # Bounding box dimension z
    diag.update({'roi_' + im_type + '_Morph_boxBoundDimZ':
                 box_bound_morph[2, 1] - box_bound_morph[2, 0] + 1})

    # Voxel number
    diag.update({'roi_' + im_type + '_Int_voxNumb': np.size(x_gl_int)})

    # Voxel number
    diag.update({'roi_' + im_type + '_Morph_voxNumb': np.size(x_gl_morph)})

    # Mean intensity
    diag.update({'roi_' + im_type + '_meanInt': np.mean(x_gl_int)})

    # Minimum intensity
    diag.update({'roi_' + im_type + '_minInt': np.min(x_gl_int)})

    # Maximum intensity
    diag.update({'roi_' + im_type + '_maxInt': np.max(x_gl_int)})

    return diag



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Tuple, Union

import numpy as np


def intrinsicToWorld(R, xIntrinsic: float, yIntrinsic: float, zIntrinsic:float) -> Tuple[float, float, float]:
    """Convert from intrinsic to world coordinates.

    Args:
        R (imref3d): imref3d object (same functionality of MATLAB imref3d class)
        xIntrinsic (float):  Coordinates along the x-dimension in the intrinsic coordinate system
        yIntrinsic (float):  Coordinates along the y-dimension in the intrinsic coordinate system
        zIntrinsic (float):  Coordinates along the z-dimension in the intrinsic coordinate system

    Returns:
        float: world coordinates
    """
    return R.intrinsicToWorld(xIntrinsic=xIntrinsic, yIntrinsic=yIntrinsic, zIntrinsic=zIntrinsic)


def worldToIntrinsic(R, xWorld: float, yWorld: float, zWorld: float) -> Tuple[float, float, float] :
    """Convert from world coordinates to intrinsic.

    Args:
        R (imref3d): imref3d object (same functionality of MATLAB imref3d class)
        xWorld (float): Coordinates along the x-dimension in the intrinsic coordinate system
        yWorld (float): Coordinates along the y-dimension in the intrinsic coordinate system
        zWorld (float): Coordinates along the z-dimension in the intrinsic coordinate system

    Returns:
        _type_: intrinsic coordinates
    """
    return R.worldToIntrinsic(xWorld=xWorld, yWorld=yWorld, zWorld=zWorld)


def sizes_match(R, A):
    """Compares whether the two imref3d objects have the same size.

    Args:
        R (imref3d): First imref3d object.
        A (imref3d): Second imref3d object.

    Returns:
        bool: True if ``R`` and ``A`` have the same size, and false if not.

    """
    return np.all(R.imageSize == A.imageSize)


class imref3d:
    """This class mirrors the functionality of the matlab imref3d class

    An `imref3d object <https://www.mathworks.com/help/images/ref/imref3d.html>`_ 
    stores the relationship between the intrinsic coordinates 
    anchored to the columns,  rows, and planes of a 3-D image and the spatial 
    location of the same column, row, and plane locations in a world coordinate system.

    The image is sampled regularly in the planar world-x, world-y, and world-z coordinates 
    of the coordinate system such that intrinsic-x, -y and -z values align with world-x, -y 
    and -z values, respectively. The resolution in each dimension can be different.

    Args:
        ImageSize (ndarray, optional): Number of elements in each spatial dimension, 
                                       specified as a 3-element positive row vector.
        PixelExtentInWorldX (float, optional): Size of a single pixel in the x-dimension 
                                               measured in the world coordinate system.
        PixelExtentInWorldY (float, optional): Size of a single pixel in the y-dimension 
                                               measured in the world coordinate system.
        PixelExtentInWorldZ (float, optional): Size of a single pixel in the z-dimension 
                                               measured in the world coordinate system.
        xWorldLimits (ndarray, optional): Limits of image in world x, specified as a 2-element row vector, 
                                          [xMin xMax].
        yWorldLimits (ndarray, optional): Limits of image in world y, specified as a 2-element row vector, 
                                          [yMin yMax].
        zWorldLimits (ndarray, optional): Limits of image in world z, specified as a 2-element row vector, 
                                          [zMin zMax].
        
    Attributes:
        ImageSize (ndarray): Number of elements in each spatial dimension, 
                             specified as a 3-element positive row vector.
        PixelExtentInWorldX (float): Size of a single pixel in the x-dimension 
                                     measured in the world coordinate system.
        PixelExtentInWorldY (float): Size of a single pixel in the y-dimension 
                                     measured in the world coordinate system.
        PixelExtentInWorldZ (float): Size of a single pixel in the z-dimension 
                                     measured in the world coordinate system.
        XIntrinsicLimits (ndarray): Limits of image in intrinsic units in the x-dimension, 
                                    specified as a 2-element row vector [xMin xMax].
        YIntrinsicLimits (ndarray): Limits of image in intrinsic units in the y-dimension, 
                                    specified as a 2-element row vector [yMin yMax].
        ZIntrinsicLimits (ndarray): Limits of image in intrinsic units in the z-dimension, 
                                    specified as a 2-element row vector [zMin zMax].
        ImageExtentInWorldX (float): Span of image in the x-dimension in 
                                     the world coordinate system.
        ImageExtentInWorldY (float): Span of image in the y-dimension in 
                                     the world coordinate system.
        ImageExtentInWorldZ (float): Span of image in the z-dimension in 
                                     the world coordinate system.
        xWorldLimits (ndarray): Limits of image in world x, specified as a 2-element row vector, 
                                [xMin xMax].
        yWorldLimits (ndarray): Limits of image in world y, specified as a 2-element row vector, 
                                [yMin yMax].
        zWorldLimits (ndarray): Limits of image in world z, specified as a 2-element row vector, 
                                [zMin zMax].
    """

    def __init__(self, 
                imageSize=None, 
                pixelExtentInWorldX=1.0,
                pixelExtentInWorldY=1.0, 
                pixelExtentInWorldZ=1.0,
                xWorldLimits=None, 
                yWorldLimits=None, 
                zWorldLimits=None) -> None:

        # Check if imageSize is an ndarray, and cast to ndarray otherwise
        self.ImageSize = self._parse_to_ndarray(x=imageSize, n=3)

        # Size of single voxels along axis in world coordinate system.
        # Equivalent to voxel spacing.
        self.PixelExtentInWorldX = pixelExtentInWorldX
        self.PixelExtentInWorldY = pixelExtentInWorldY
        self.PixelExtentInWorldZ = pixelExtentInWorldZ

        # Limits of the image in intrinsic coordinates
        # AZ: this differs from DICOM, which assumes that the origin lies
        # at the center of the first voxel.
        if imageSize is not None:
            self.XIntrinsicLimits = np.array([-0.5, imageSize[0]-0.5])
            self.YIntrinsicLimits = np.array([-0.5, imageSize[1]-0.5])
            self.ZIntrinsicLimits = np.array([-0.5, imageSize[2]-0.5])
        else:
            self.XIntrinsicLimits = None
            self.YIntrinsicLimits = None
            self.ZIntrinsicLimits = None

        # Size of the image in world coordinates
        if imageSize is not None:
            self.ImageExtentInWorldX = imageSize[0] * pixelExtentInWorldX
            self.ImageExtentInWorldY = imageSize[1] * pixelExtentInWorldY
            self.ImageExtentInWorldZ = imageSize[2] * pixelExtentInWorldZ
        else:
            self.ImageExtentInWorldX = None
            self.ImageExtentInWorldY = None
            self.ImageExtentInWorldZ = None

        # Limits of the image in the world coordinates
        self.XWorldLimits = self._parse_to_ndarray(x=xWorldLimits, n=2)
        self.YWorldLimits = self._parse_to_ndarray(x=yWorldLimits, n=2)
        self.ZWorldLimits = self._parse_to_ndarray(x=zWorldLimits, n=2)

        if xWorldLimits is None and imageSize is not None:
            self.XWorldLimits = np.array([0.0, self.ImageExtentInWorldX])
        if yWorldLimits is None and imageSize is not None:
            self.YWorldLimits = np.array([0.0, self.ImageExtentInWorldY])
        if zWorldLimits is None and imageSize is not None:
            self.ZWorldLimits = np.array([0.0, self.ImageExtentInWorldZ])

    def _parse_to_ndarray(self,
                          x: np.iterable,
                          n=None) -> np.ndarray:
        """Internal function to cast input to a numpy array.

        Args:
            x (iterable): Object that supports __iter__.
            n (int, optional): expected length.

        Returns:
            ndarray: iterable input as a numpy array.
        """
        if x is not None:
            # Cast to ndarray
            if not isinstance(x, np.ndarray):
                x = np.array(x)

            # Check length
            if n is not None:
                if not len(x) == n:
                    raise ValueError(
                        "Length of array does not meet the expected length.", len(x), n)

        return x

    def intrinsicToWorld(self, 
                         xIntrinsic: np.ndarray, 
                         yIntrinsic: np.ndarray, 
                         zIntrinsic: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Convert from intrinsic to world coordinates.

        Args:
            xIntrinsic (ndarray): Coordinates along the x-dimension in the intrinsic coordinate system.
            yIntrinsic (ndarray): Coordinates along the y-dimension in the intrinsic coordinate system.
            zIntrinsic (ndarray): Coordinates along the z-dimension in the intrinsic coordinate system.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: [xWorld, yWorld, zWorld] in world coordinate system.
        """
        xWorld = (self.XWorldLimits[0] + 0.5*self.PixelExtentInWorldX) + \
            xIntrinsic * self.PixelExtentInWorldX
        yWorld = (self.YWorldLimits[0] + 0.5*self.PixelExtentInWorldY) + \
            yIntrinsic * self.PixelExtentInWorldY
        zWorld = (self.ZWorldLimits[0] + 0.5*self.PixelExtentInWorldZ) + \
            zIntrinsic * self.PixelExtentInWorldZ

        return xWorld, yWorld, zWorld

    def worldToIntrinsic(self, 
                         xWorld: np.ndarray, 
                         yWorld: np.ndarray, 
                         zWorld: np.ndarray)-> Union[np.ndarray,
                                                     np.ndarray,
                                                     np.ndarray]:
        """Converts from world coordinates to intrinsic coordinates.

        Args:
            xWorld (ndarray): Coordinates along the x-dimension in the world coordinate system.
            yWorld (ndarray): Coordinates along the y-dimension in the world coordinate system.
            zWorld (ndarray): Coordinates along the z-dimension in the world coordinate system.

        Returns:
            ndarray: [xIntrinsic,yIntrinsic,zIntrinsic] in intrinsic coordinate system.
        """

        xIntrinsic = (
            xWorld - (self.XWorldLimits[0] + 0.5*self.PixelExtentInWorldX)) / self.PixelExtentInWorldX
        yIntrinsic = (
            yWorld - (self.YWorldLimits[0] + 0.5*self.PixelExtentInWorldY)) / self.PixelExtentInWorldY
        zIntrinsic = (
            zWorld - (self.ZWorldLimits[0] + 0.5*self.PixelExtentInWorldZ)) / self.PixelExtentInWorldZ

        return xIntrinsic, yIntrinsic, zIntrinsic

    def contains_point(self,
                       xWorld: np.ndarray,
                       yWorld: np.ndarray,
                       zWorld: np.ndarray) -> np.ndarray:
        """Determines which points defined by ``xWorld``, ``yWorld`` and ``zWorld``.

        Args:
            xWorld (ndarray): Coordinates along the x-dimension in the world coordinate system.
            yWorld (ndarray): Coordinates along the y-dimension in the world coordinate system.
            zWorld (ndarray): Coordinates along the z-dimension in the world coordinate system.

        Returns:
            ndarray: boolean array for coordinate sets that are within the bounds of the image.
        """
        xInside = np.logical_and(
            xWorld >= self.XWorldLimits[0], xWorld <= self.XWorldLimits[1])
        yInside = np.logical_and(
            yWorld >= self.YWorldLimits[0], yWorld <= self.YWorldLimits[1])
        zInside = np.logical_and(
            zWorld >= self.ZWorldLimits[0], zWorld <= self.ZWorldLimits[1])

        return xInside + yInside + zInside == 3

    def WorldLimits(self,
                    axis=None,
                    newValue=None) -> Union[np.ndarray, None]:
        """Sets the WorldLimits to the new value for the given ``axis``.
        If the newValue is None, the method returns the attribute value.

        Args:
            axis (str, optional): Specify the dimension, must be 'X', 'Y' or 'Z'.
            newValue (iterable, optional): New value for the WorldLimits attribute.

        Returns:
            ndarray: Limits of image in world along the axis-dimension.
        """
        if newValue is None:
            # Get value
            if axis == "X":
                return self.XWorldLimits
            elif axis == "Y":
                return self.YWorldLimits
            elif axis == "Z":
                return self.ZWorldLimits
        else:
            # Set value
            if axis == "X":
                self.XWorldLimits = self._parse_to_ndarray(x=newValue, n=2)
            elif axis == "Y":
                self.YWorldLimits = self._parse_to_ndarray(x=newValue, n=2)
            elif axis == "Z":
                self.ZWorldLimits = self._parse_to_ndarray(x=newValue, n=2)

    def PixelExtentInWorld(self, axis=None) -> Union[float, None]:
        """Returns the PixelExtentInWorld attribute value for the given ``axis``.

        Args:
            axis (str, optional): Specify the dimension, must be 'X', 'Y' or 'Z'.

        Returns:
            float: Size of a single pixel in the axis-dimension measured in the world coordinate system.
        """
        if axis == "X":
            return self.PixelExtentInWorldX
        elif axis == "Y":
            return self.PixelExtentInWorldY
        elif axis == "Z":
            return self.PixelExtentInWorldZ

    def IntrinsicLimits(self,
                        axis=None) -> Union[np.ndarray,
                                            None]:
        """Returns the IntrinsicLimits attribute value for the given ``axis``.

        Args:
            axis (str, optional): Specify the dimension, must be 'X', 'Y' or 'Z'.

        Returns:
            ndarray: Limits of image in intrinsic units in the axis-dimension, specified as a 2-element row vector [xMin xMax].
        """
        if axis == "X":
            return self.XIntrinsicLimits
        elif axis == "Y":
            return self.YIntrinsicLimits
        elif axis == "Z":
            return self.ZIntrinsicLimits

    def ImageExtentInWorld(self,
                           axis=None) -> Union[float,
                                               None]:
        """Returns the ImageExtentInWorld attribute value for the given ``axis``.

        Args:
            axis (str, optional): Specify the dimension, must be 'X', 'Y' or 'Z'.

        Returns:
            ndarray: Span of image in the axis-dimension in the world coordinate system.

        """
        if axis == "X":
            return self.ImageExtentInWorldX
        elif axis == "Y":
            return self.ImageExtentInWorldY
        elif axis == "Z":
            return self.ImageExtentInWorldZ



from copy import deepcopy
from typing import Union

import numpy as np
try:
    import pycuda.autoinit
    import pycuda.driver as cuda
    from pycuda.autoinit import context
    from pycuda.compiler import SourceModule
except Exception as e:
    print("PyCUDA is not installed. Please install it to use the textural filters.", e)
    import_failed = True

from ..processing.discretisation import discretize
from .textural_filters_kernels import glcm_kernel, single_glcm_kernel


class TexturalFilter():
    """The Textural filter class. This class is used to apply textural filters to an image. The textural filters are
    chosen from the following families: GLCM, NGTDM, GLDZM, GLSZM, NGLDM, GLRLM. The computation is done using CUDA."""

    def __init__(
                self,
                family: str,
                size: int = 3,
                local: bool = False
                ):

        """
        The constructor for the textural filter class.

        Args:
            family (str): The family of the textural filter.
            size (int, optional): The size of the kernel, which will define the filter kernel dimension.
            local (bool, optional): If true, the discrete will be computed locally, else globally.
        
        Returns:
            None.
        """

        assert size % 2 == 1 and size > 0, "size should be a positive odd number."
        assert isinstance(family, str) and family.upper() in ["GLCM", "NGTDM", "GLDZM", "GLSZM", "NGLDM", "GLRLM"],\
            "family should be a string and should be one of the following: GLCM, NGTDM, GLDZM, GLSZM, NGLDM, GLRLM."

        self.family = family
        self.size = size
        self.local = local
        self.glcm_features = [
            "Fcm_joint_max",
            "Fcm_joint_avg",
            "Fcm_joint_var",
            "Fcm_joint_entr",
            "Fcm_diff_avg",
            "Fcm_diff_var",
            "Fcm_diff_entr",
            "Fcm_sum_avg",
            "Fcm_sum_var",
            "Fcm_sum_entr",
            "Fcm_energy",
            "Fcm_contrast",
            "Fcm_dissimilarity",
            "Fcm_inv_diff",
            "Fcm_inv_diff_norm",
            "Fcm_inv_diff_mom",
            "Fcm_inv_diff_mom_norm",
            "Fcm_inv_var",
            "Fcm_corr",
            "Fcm_auto_corr",
            "Fcm_clust_tend",
            "Fcm_clust_shade",
            "Fcm_clust_prom",
            "Fcm_info_corr1",
            "Fcm_info_corr2"
        ]

    def __glcm_filter(
            self,
            input_images: np.ndarray,
            discretization : dict,
            user_set_min_val: float,
            feature = None
        ) -> np.ndarray:
        """
        Apply a textural filter to the input image.

        Args:
            input_images (ndarray): The images to filter.
            discretization (dict): The discretization parameters.
            user_set_min_val (float): The minimum value to use for the discretization.
            family (str, optional): The family of the textural filter.
            feature (str, optional): The feature to extract from the family. if not specified, all the features of the
                family will be extracted.
        
        Returns:
            ndarray: The filtered image.
        """

        if feature:
            if isinstance(feature, str):
                assert feature in self.glcm_features,\
                    "feature should be a string or an integer and should be one of the following: " + ", ".join(self.glcm_features) + "."
            elif isinstance(feature, int):
                assert feature in range(len(self.glcm_features)),\
                    "feature's index should be an integer between 0 and " + str(len(self.glcm_features) - 1) + "."
            else:
                raise TypeError("feature should be an integer or a string from the following list: " + ", ".join(self.glcm_features) + ".")
                

        # Pre-processing of the input volume
        padding_size = (self.size - 1) // 2
        input_images = np.pad(input_images[:, :, :], padding_size, mode="constant", constant_values=np.nan)
        input_images_copy = deepcopy(input_images)

        # Set up the strides
        strides = (
            input_images_copy.shape[2] * input_images_copy.shape[1] * input_images_copy.dtype.itemsize,
            input_images_copy.shape[2] * input_images_copy.dtype.itemsize,
            input_images_copy.dtype.itemsize
        )
        input_images = np.lib.stride_tricks.as_strided(input_images, shape=input_images.shape, strides=strides)
        input_images[:,:,:] = input_images_copy[:, :, :]

        if self.local:
            # Discretization (to get the global max value)
            if discretization['type'] == "FBS":
                print("Warning: FBS local discretization is equivalent to global discretization.")
                n_q = discretization['bw']
            elif discretization['type'] == "FBN" and discretization['adapted']:
                n_q = (np.nanmax(input_images) - np.nanmin(input_images)) // discretization['bw']
                user_set_min_val = np.nanmin(input_images)
            elif discretization['type'] == "FBN":
                n_q = discretization['bn']
                user_set_min_val = np.nanmin(input_images)
            else:
                raise ValueError("Discretization should be either FBS or FBN.")
            
            temp_vol, _ = discretize(
                vol_re=input_images,
                discr_type=discretization['type'],
                n_q=n_q,
                user_set_min_val=user_set_min_val,
                ivh=False
            )

            # Initialize the filtering parameters
            max_vol = np.nanmax(temp_vol)

            del temp_vol
        
        else:
            # Discretization
            if discretization['type'] == "FBS":
                n_q = discretization['bw']
            elif discretization['type'] == "FBN":
                n_q = discretization['bn']
                user_set_min_val = np.nanmin(input_images)
            else:
                raise ValueError("Discretization should be either FBS or FBN.")
            
            input_images, _ = discretize(
                vol_re=input_images,
                discr_type=discretization['type'],
                n_q=n_q,
                user_set_min_val=user_set_min_val,
                ivh=False
            )

            # Initialize the filtering parameters
            max_vol = np.nanmax(input_images)

        volume_copy = deepcopy(input_images)

        # Filtering
        if feature is not None:
            # Select the feature to compute
            feature = self.glcm_features.index(feature) if isinstance(feature, str) else feature

            # Initialize the kernel
            kernel_glcm = single_glcm_kernel.substitute(
                max_vol=int(max_vol),
                filter_size=self.size,
                shape_volume_0=int(volume_copy.shape[0]),
                shape_volume_1=int(volume_copy.shape[1]),
                shape_volume_2=int(volume_copy.shape[2]),
                discr_type=discretization['type'],
                n_q=n_q,
                min_val=user_set_min_val,
                feature_index=feature
            )

        else:
            # Create the final volume to store the results
            input_images = np.zeros((input_images.shape[0], input_images.shape[1], input_images.shape[2], 25), dtype=np.float32)

            # Fill with nan
            input_images[:] = np.nan

            # Initialize the kernel
            kernel_glcm = glcm_kernel.substitute(
                max_vol=int(max_vol),
                filter_size=self.size,
                shape_volume_0=int(volume_copy.shape[0]),
                shape_volume_1=int(volume_copy.shape[1]),
                shape_volume_2=int(volume_copy.shape[2]),
                discr_type=discretization['type'],
                n_q=n_q,
                min_val=user_set_min_val
            )

        # Compile the CUDA kernel
        if not import_failed:
            mod = SourceModule(kernel_glcm, no_extern_c=True)
            if self.local:
                process_loop_kernel = mod.get_function("glcm_filter_local")
            else:
                process_loop_kernel = mod.get_function("glcm_filter_global")

            # Allocate GPU memory
            volume_gpu = cuda.mem_alloc(input_images.nbytes)
            volume_gpu_copy = cuda.mem_alloc(volume_copy.nbytes)

            # Copy data to the GPU
            cuda.memcpy_htod(volume_gpu, input_images)
            cuda.memcpy_htod(volume_gpu_copy, volume_copy)

            # Set up the grid and block dimensions
            block_dim = (16, 16, 1)  # threads per block
            grid_dim = (
                int((volume_copy.shape[0] - 1) // block_dim[0] + 1),
                int((volume_copy.shape[1] - 1) // block_dim[1] + 1),
                int((volume_copy.shape[2] - 1) // block_dim[2] + 1)
            )   # blocks in the grid

            # Run the kernel
            process_loop_kernel(volume_gpu, volume_gpu_copy, block=block_dim, grid=grid_dim)

            # Synchronize to ensure all CUDA operations are complete
            context.synchronize()

            # Copy data back to the CPU
            cuda.memcpy_dtoh(input_images, volume_gpu)

            # Free the allocated GPU memory
            volume_gpu.free()
            volume_gpu_copy.free()
            del volume_copy

            # unpad the volume
            if feature: # 3D (single-feature)
                input_images = input_images[padding_size:-padding_size, padding_size:-padding_size, padding_size:-padding_size]
            else: # 4D (all features)
                input_images = input_images[padding_size:-padding_size, padding_size:-padding_size, padding_size:-padding_size, :]

            return input_images

        else:
            return None
    
    def __call__(
            self,
            input_images: np.ndarray,
            discretization : dict,
            user_set_min_val: float,
            family: str = "GLCM",
            feature : str = None,
            size: int = None,
            local: bool = False
        ) -> np.ndarray:
        """
        Apply a textural filter to the input image.

        Args:
            input_images (ndarray): The images to filter.
            discretization (dict): The discretization parameters.
            user_set_min_val (float): The minimum value to use for the discretization.
            family (str, optional): The family of the textural filter.
            feature (str, optional): The feature to extract from the family. if not specified, all the features of the
                family will be extracted.
            size (int, optional): The filter size.
            local (bool, optional): If true, the discretization will be computed locally, else globally.
        
        Returns:
            ndarray: The filtered image.
        """
        # Initialization
        if family:
            self.family = family
        if size:
            self.size = size
        if local:
            self.local = local

        # Filtering
        if self.family.lower() == "glcm":
            filtered_images = self.__glcm_filter(input_images, discretization, user_set_min_val, feature)
        else:
            raise NotImplementedError("Only GLCM is implemented for now.")

        return filtered_images
